"""
Tests of the flitter language virtual machine

Note that `Import` is currently not tested.
"""

import gc
import tracemalloc
import unittest
import unittest.mock

from flitter.model import Vector, Node, Context, StateDict, null, true, false
from flitter.language.vm import Program, Function, VectorStack


class TestBasicInstructions(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)

    def test_Add(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.add()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [7])

    def test_Append(self):
        node = Node('foo', {'test'}, {'x': Vector(1)})
        self.program.literal(node)
        self.program.literal(Node('bar'))
        self.program.append()
        self.program.literal(Vector([Node('baz1', None, {'a': Vector(1)}), Node('baz2', None, {'b': Vector(2)})]))
        self.program.literal(Vector(Node('baz3', None, {'c': Vector(3)})))
        self.program.append(2)
        stack = self.program.execute(self.context)
        self.assertEqual(len(stack), 1)
        self.assertEqual(repr(stack[0]), "(!foo #test x=1\n !bar\n !baz1 a=1\n !baz2 b=2\n !baz3 c=3)")

    def test_Attribute(self):
        foo = Node('foo', {'test'}, {'x': Vector(1)})
        bar = Node('bar', None, {'x': Vector(2)})
        self.names['bar'] = Vector(bar)
        self.program.literal(foo)
        self.program.literal(bar)
        self.program.local_push(1)
        self.program.literal(12)
        self.program.attribute('y')
        self.program.literal(5)
        self.program.attribute('x')
        self.program.literal(null)
        self.program.attribute('y')
        self.program.local_load(0)
        self.program.literal(7)
        self.program.attribute('y')
        stack = self.program.execute(self.context)
        self.assertEqual(len(stack), 2)
        self.assertEqual(repr(stack[0]), "(!foo #test x=5)")
        self.assertFalse(stack[0][0] is foo)
        self.assertEqual(repr(stack[1]), "(!bar x=2 y=7)")
        self.assertFalse(stack[1][0] is bar)
        self.assertEqual(foo['x'], Vector(1))

    def test_Compose(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.compose(2)
        self.program.literal(Vector(['a', 'b']))
        self.program.literal(Vector('c'))
        self.program.compose(2)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [Vector([3, 4]), Vector(['a', 'b', 'c'])])

    def test_Drop(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.drop()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [3])

    def test_Dup(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.dup()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [3, 4, 4])

    def test_Eq(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.eq()
        self.program.literal(5)
        self.program.literal(5)
        self.program.eq()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true])

    def test_Ceil(self):
        self.program.literal(-3.5)
        self.program.ceil()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [-3])

    def test_Floor(self):
        self.program.literal(-3.5)
        self.program.floor()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [-4])

    def test_FloorDiv(self):
        self.program.literal(11)
        self.program.literal(4)
        self.program.floordiv()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [2])

    def test_Fract(self):
        self.program.literal(-3.5)
        self.program.fract()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [0.5])

    def test_Ge(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.ge()
        self.program.literal(3)
        self.program.literal(3)
        self.program.ge()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true])

    def test_Gt(self):
        self.program.literal(3)
        self.program.literal(3)
        self.program.gt()
        self.program.literal(4)
        self.program.literal(3)
        self.program.gt()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true])

    def test_IndexLiteral(self):
        self.program.literal(range(1, 11))
        self.program.slice_literal(Vector(3))
        self.assertEqual(str(self.program.instructions[-1]), 'IndexLiteral 3')
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [4])

    def test_LiteralNode(self):
        self.program.literal(Node('foo', {'test'}, {'x': Vector(1)}))
        self.assertEqual(str(self.program.instructions[-1]), 'LiteralNode !foo #test x=1')

    def test_LiteralNodes(self):
        self.program.literal([Node('foo', {'test'}, {'x': Vector(1)}), Node('bar')])
        self.assertEqual(str(self.program.instructions[-1]), 'LiteralNodes (!foo #test x=1, !bar)')

    def test_Literal(self):
        self.program.literal(range(10))
        self.assertEqual(str(self.program.instructions[-1]), 'Literal 0;1;2;3;4;5;6;7;8;9')

    def test_Le(self):
        self.program.literal(3)
        self.program.literal(3)
        self.program.le()
        self.program.literal(4)
        self.program.literal(3)
        self.program.le()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [true, false])

    def test_StoreGlobal(self):
        self.program.literal((1, 2))
        self.program.store_global('x')
        self.program.literal(3)
        self.program.store_global('y')
        self.program.literal("Hello world!")
        self.program.store_global('z')
        stack = self.program.execute(self.context)
        self.assertEqual(len(stack), 0)
        self.assertEqual(self.names, {'x': Vector([1, 2]), 'y': Vector(3), 'z': Vector("Hello world!")})

    def test_Lookup(self):
        self.state['y'] = 12
        self.program.literal('x')
        self.program.lookup()
        self.program.literal('y')
        self.program.lookup()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [null, 12])

    def test_LookupLiteral(self):
        self.state['y'] = 12
        self.program.lookup_literal(Vector('x'))
        self.program.lookup_literal(Vector('y'))
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [null, 12])

    def test_Lt(self):
        self.program.literal(3)
        self.program.literal(3)
        self.program.lt()
        self.program.literal(3)
        self.program.literal(4)
        self.program.lt()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true])

    def test_Mod(self):
        self.program.literal(7)
        self.program.literal(4)
        self.program.mod()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [3])

    def test_Mul(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.mul()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [12])

    def test_Ne(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.ne()
        self.program.literal(5)
        self.program.literal(5)
        self.program.ne()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [true, false])

    def test_Neg(self):
        self.program.literal(3)
        self.program.neg()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [-3])

    def test_Not(self):
        self.program.literal(true)
        self.program.not_()
        self.program.literal(false)
        self.program.not_()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true])

    def test_Pos(self):
        self.program.literal(3)
        self.program.pos()
        self.program.literal('hello')
        self.program.pos()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [3, null])

    def test_Pow(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.pow()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [81])

    def test_Pragma(self):
        self.program.literal(3)
        self.program.pragma('x')
        stack = self.program.execute(self.context)
        self.assertEqual(len(stack), 0)
        self.assertEqual(self.context.pragmas, {'x': Vector(3)})

    def test_Range(self):
        self.program.literal(0)
        self.program.literal(10)
        self.program.literal(2)
        self.program.range()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [(0, 2, 4, 6, 8)])

    def test_Slice(self):
        self.program.literal(range(1, 11))
        self.program.literal((0, 2))
        self.program.slice()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [(1, 3)])

    def test_SliceLiteral(self):
        self.program.literal(range(1, 11))
        self.program.slice_literal(Vector([0, 2]))
        self.assertEqual(str(self.program.instructions[-1]), 'SliceLiteral 0;2')
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [(1, 3)])

    def test_Sub(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.sub()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [-1])

    def test_Tag(self):
        node = Node('foo', None, {'x': Vector(1)})
        self.program.literal(node)
        self.program.tag('test')
        stack = self.program.execute(self.context)
        self.assertEqual(len(stack), 1)
        self.assertEqual(repr(stack[0]), "(!foo #test x=1)")

    def test_TrueDiv(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.truediv()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [0.75])

    def test_Xor(self):
        self.program.literal(false)
        self.program.literal(false)
        self.program.xor()
        self.program.literal(true)
        self.program.literal(false)
        self.program.xor()
        self.program.literal(false)
        self.program.literal(true)
        self.program.xor()
        self.program.literal(true)
        self.program.literal(true)
        self.program.xor()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [false, true, true, false])


class TestJumps(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)

    def test_jump(self):
        LABEL = self.program.new_label()
        self.program.literal(1)
        self.program.jump(LABEL)
        self.program.literal(2)
        self.program.label(LABEL)
        self.program.literal(3)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [1, 3])

    def test_branch_true(self):
        LABEL1 = self.program.new_label()
        LABEL2 = self.program.new_label()
        self.program.literal(1)
        self.program.literal(true)
        self.program.branch_true(LABEL1)
        self.program.literal(2)
        self.program.label(LABEL1)
        self.program.literal(3)
        self.program.literal(false)
        self.program.branch_true(LABEL2)
        self.program.literal(4)
        self.program.label(LABEL2)
        self.program.literal(5)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [1, 3, 4, 5])

    def test_branch_false(self):
        LABEL1 = self.program.new_label()
        LABEL2 = self.program.new_label()
        self.program.literal(1)
        self.program.literal(true)
        self.program.branch_false(LABEL1)
        self.program.literal(2)
        self.program.label(LABEL1)
        self.program.literal(3)
        self.program.literal(false)
        self.program.branch_false(LABEL2)
        self.program.literal(4)
        self.program.label(LABEL2)
        self.program.literal(5)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [1, 2, 3, 5])


class TestForLoops(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)

    def test_simple(self):
        NEXT = self.program.new_label()
        END = self.program.new_label()
        self.program.literal(Vector.range(10))
        self.program.begin_for()
        self.program.literal(null)
        self.program.local_push(1)
        self.program.label(NEXT)
        self.program.next(1, END)
        self.program.local_load(0)
        self.program.literal(2)
        self.program.mul()
        self.program.jump(NEXT)
        self.program.label(END)
        self.program.end_for_compose()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [Vector.range(0, 20, 2)])

    def test_multiple_names(self):
        NEXT = self.program.new_label()
        END = self.program.new_label()
        self.program.literal(Vector.range(10))
        self.program.begin_for()
        self.program.literal(null)
        self.program.local_push(2)
        self.program.label(NEXT)
        self.program.next(2, END)
        self.program.local_load(1)
        self.program.literal(2)
        self.program.mul()
        self.program.local_load(0)
        self.program.compose(2)
        self.program.jump(NEXT)
        self.program.label(END)
        self.program.end_for_compose()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [(0, 1, 4, 3, 8, 5, 12, 7, 16, 9)])

    def test_short(self):
        NEXT = self.program.new_label()
        END = self.program.new_label()
        self.program.literal(Vector.range(9))
        self.program.begin_for()
        self.program.literal(null)
        self.program.local_push(2)
        self.program.label(NEXT)
        self.program.next(2, END)
        self.program.local_load(1)
        self.program.literal(2)
        self.program.mul()
        self.program.local_load(0)
        self.program.compose(2)
        self.program.jump(NEXT)
        self.program.label(END)
        self.program.end_for_compose()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [(0, 1, 4, 3, 8, 5, 12, 7, 16)])

    def test_empty(self):
        NEXT = self.program.new_label()
        END = self.program.new_label()
        self.program.literal(null)
        self.program.begin_for()
        self.program.literal(null)
        self.program.local_push(1)
        self.program.label(NEXT)
        self.program.next(1, END)
        self.program.local_load(0)
        self.program.literal(2)
        self.program.mul()
        self.program.jump(NEXT)
        self.program.label(END)
        self.program.end_for_compose()
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [null])


class TestLocalVars(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)

    def test_simple(self):
        self.program.literal(5)
        self.program.local_push(1)
        self.program.literal(2)
        self.program.local_load(0)
        self.program.add()
        self.program.local_drop(1)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [7])

    def test_shadow(self):
        self.program.literal(5)
        self.program.local_push(1)
        self.program.literal(2)
        self.program.local_load(0)
        self.program.add()
        self.program.local_push(1)
        self.program.local_load(0)
        self.program.dup()
        self.program.local_drop(1)
        self.program.local_load(0)
        self.program.add()
        self.program.local_drop(1)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [7, 12])

    def test_multiple(self):
        self.program.literal([1, 2, 3])
        self.program.local_push(2)
        self.program.local_load(0)
        self.program.dup()
        self.program.local_load(1)
        self.program.add()
        self.program.local_drop(2)
        self.program.literal([1, 2])
        self.program.local_push(3)
        self.program.local_load(0)
        self.program.local_drop(3)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [2, 3, 1])


class TestFunc(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)
        self.program.literal(2)
        self.program.local_push(1)
        self.program.local_load(0)
        self.program.literal(null)
        self.program.literal(1)
        END = self.program.new_label()
        self.program.jump(END)
        START = self.program.new_label()
        self.program.label(START)
        self.program.local_load(1)
        self.program.local_load(0)
        self.program.add()
        self.program.local_load(3)
        self.program.add()
        LABEL1 = self.program.new_label()
        self.program.literal('test')
        self.program.lookup()
        self.program.branch_false(LABEL1)
        self.program.literal(1)
        self.program.add()
        self.program.label(LABEL1)
        self.program.exit()
        self.program.label(END)
        self.program.func(START, 'f', ('x', 'y'), 1)
        self.program.local_push(1)

    def test_declare(self):
        lnames = []
        stack = self.program.execute(self.context, lnames=lnames)
        self.assertEqual(len(stack), 0)
        self.assertEqual(len(lnames), 2)
        self.assertEqual(lnames[0], 2)
        function, = lnames[1]
        self.assertTrue(isinstance(function, Function))
        self.assertEqual(function.__name__, 'f')
        self.assertEqual(function.parameters, ('x', 'y'))
        self.assertEqual(function.defaults, (null, 1))
        self.assertIs(function.program, self.program)

    def test_call(self):
        self.program.literal(3)
        self.program.literal(4)
        self.program.local_load(0)
        self.program.call(2)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [9])

    def test_additional_lnames(self):
        self.program.literal(-2)
        self.program.local_push(1)
        self.program.literal(3)
        self.program.literal(4)
        self.program.local_load(1)
        self.program.call(2)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [9])

    def test_default_arg(self):
        self.program.literal(3)
        self.program.local_load(0)
        self.program.call(1)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [6])

    def test_default_arg2(self):
        self.program.local_load(0)
        self.program.call(0)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [null])

    def test_state(self):
        self.state['test'] = Vector(1)
        self.program.literal(3)
        self.program.literal(4)
        self.program.local_load(0)
        self.program.call(2)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [10])


class TestRecursiveFunc(unittest.TestCase):
    def test_recursive_fib(self):
        self.program = Program()
        self.state = StateDict()
        self.names = {}
        self.context = Context(state=self.state, names=self.names)
        self.program.literal(null)
        END = self.program.new_label()
        self.program.jump(END)
        START = self.program.new_label()
        self.program.label(START)
        self.program.local_load(0)
        self.program.literal(2)
        self.program.lt()
        LABEL2 = self.program.new_label()
        self.program.branch_false(LABEL2)
        self.program.local_load(0)
        LABEL1 = self.program.new_label()
        self.program.jump(LABEL1)
        self.program.label(LABEL2)
        self.program.local_load(0)
        self.program.literal(1)
        self.program.sub()
        self.program.local_load(1)
        self.program.call(1)
        self.program.local_load(0)
        self.program.literal(2)
        self.program.sub()
        self.program.local_load(1)
        self.program.call(1)
        self.program.add()
        self.program.label(LABEL1)
        self.program.exit()
        self.program.label(END)
        self.program.func(START, 'fib', ('x',))
        self.program.local_push(1)
        self.program.literal(10)
        self.program.local_load(0)
        self.program.call(1)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [55])


class TestCalls(unittest.TestCase):
    def setUp(self):
        self.program = Program()
        self.state = StateDict()
        self.test_function = unittest.mock.Mock()
        del self.test_function.context_func
        self.test_function.return_value = Vector(12)
        self.context_function = unittest.mock.Mock(context_func=True)
        self.context_function.return_value = Vector(12)
        self.names = {'test': Vector(self.test_function), 'context': Vector(self.context_function)}
        self.context = Context(state=self.state, names=self.names)

    def test_no_args(self):
        self.program.local_load(0)
        self.program.call(0)
        stack = self.program.execute(self.context, lnames=[self.test_function])
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with()

    def test_one_arg(self):
        self.program.literal(1)
        self.program.local_load(0)
        self.program.call(1)
        stack = self.program.execute(self.context, lnames=[self.test_function])
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(Vector(1))

    def test_multiple_args(self):
        self.program.literal(1)
        self.program.literal(2)
        self.program.local_load(0)
        self.program.call(2)
        stack = self.program.execute(self.context, lnames=[self.test_function])
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(Vector(1), Vector(2))

    def test_kwargs(self):
        self.program.literal(1)
        self.program.literal(2)
        self.program.local_load(0)
        self.program.call(0, ('x', 'y'))
        stack = self.program.execute(self.context, lnames=[self.test_function])
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(x=Vector(1), y=Vector(2))

    def test_args_kwargs(self):
        self.program.literal(1)
        self.program.literal(2)
        self.program.literal(3)
        self.program.local_load(0)
        self.program.call(1, ('x', 'y'))
        stack = self.program.execute(self.context, lnames=[self.test_function])
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(Vector(1), x=Vector(2), y=Vector(3))

    def test_context_func(self):
        self.program.literal(1)
        self.program.literal(2)
        self.program.literal(3)
        self.program.local_load(0)
        self.program.call(1, ('x', 'y'))
        stack = self.program.execute(self.context, lnames=[self.context_function])
        self.assertEqual(stack, [12])
        self.context_function.assert_called_once_with(self.context, Vector(1), x=Vector(2), y=Vector(3))

    def test_call_fast_multiple_args(self):
        self.program.literal(1)
        self.program.literal(2)
        self.program.call_fast(self.test_function, 2)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(Vector(1), Vector(2))

    def test_call_fast_single_arg(self):
        self.program.literal(1)
        self.program.call_fast(self.test_function, 1)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with(Vector(1))

    def test_call_fast_no_args(self):
        self.program.call_fast(self.test_function, 0)
        stack = self.program.execute(self.context)
        self.assertEqual(stack, [12])
        self.test_function.assert_called_once_with()


class TestStack(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        tracemalloc.start()

    @classmethod
    def tearDownClass(cls):
        tracemalloc.stop()

    def setUp(self):
        gc.collect()
        gc.disable()

    def tearDown(self):
        for obj in gc.get_objects(0):
            self.assertNotIsInstance(obj, Vector, "Memory leak")
        gc.enable()

    def test_create(self):
        stack = VectorStack(10)
        self.assertEqual(stack.size, 10)
        self.assertEqual(len(stack), 0)

    def test_basic_push_and_pop(self):
        stack = VectorStack()
        stack.push(Vector(0))
        self.assertEqual(len(stack), 1)
        stack.push(Vector(1))
        self.assertEqual(len(stack), 2)
        self.assertEqual(stack.pop(), Vector(1))
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)

    def test_drop(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        stack.push(Vector(3))
        self.assertEqual(len(stack), 4)
        stack.drop()
        self.assertEqual(len(stack), 3)
        stack.drop(2)
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)

    def test_realloc(self):
        stack = VectorStack(1)
        stack.push(Vector(0))
        self.assertEqual(stack.size, 1)
        self.assertEqual(len(stack), 1)
        stack.push(Vector(1))
        self.assertEqual(stack.size, 2)
        self.assertEqual(len(stack), 2)
        stack.push(Vector(2))
        self.assertEqual(stack.size, 4)
        self.assertEqual(len(stack), 3)
        stack.push(Vector(3))
        self.assertEqual(stack.size, 4)
        self.assertEqual(len(stack), 4)
        stack.push(Vector(4))
        self.assertEqual(stack.size, 8)
        self.assertEqual(len(stack), 5)

    def test_peek_and_poke(self):
        stack = VectorStack()
        stack.push(Vector(0))
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.peek(), Vector(0))
        stack.poke(Vector(1))
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(1))
        self.assertEqual(len(stack), 0)

    def test_peek_at_and_poke_at(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.peek_at(0), Vector(2))
        self.assertEqual(stack.peek_at(1), Vector(1))
        self.assertEqual(stack.peek_at(2), Vector(0))
        stack.poke_at(0, Vector(3))
        stack.poke_at(1, Vector(4))
        stack.poke_at(2, Vector(5))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.pop(), Vector(3))
        self.assertEqual(stack.pop(), Vector(4))
        self.assertEqual(stack.pop(), Vector(5))
        self.assertEqual(len(stack), 0)

    def test_pop_tuple(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.pop_tuple(2), (Vector(1), Vector(2)))
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)

    def test_pop_list(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.pop_list(2), [Vector(1), Vector(2)])
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)

    def test_pop_dict(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.pop_dict(('x', 'y')), {'x': Vector(1), 'y': Vector(2)})
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)

    def test_pop_composed(self):
        stack = VectorStack()
        stack.push(Vector(0))
        stack.push(Vector(1))
        stack.push(Vector(2))
        self.assertEqual(len(stack), 3)
        self.assertEqual(stack.pop_composed(2), Vector([1, 2]))
        self.assertEqual(len(stack), 1)
        self.assertEqual(stack.pop(), Vector(0))
        self.assertEqual(len(stack), 0)



"""
Record scene node
"""

from . import ProgramNode
from ...cache import SharedCache
from .glconstants import GL_SRGB8, GL_SRGB8_ALPHA8
from .glsl import TemplateLoader


class Record(ProgramNode):
    DEFAULT_VERTEX_SOURCE = TemplateLoader.get_template('video.vert')

    def __init__(self, glctx):
        super().__init__(glctx)
        self._framebuffer = None
        self._texture = None
        self._has_alpha = None

    @property
    def texture(self):
        return self.children[0].texture if len(self.children) == 1 else None

    @property
    def framebuffer(self):
        return self._framebuffer

    def release(self):
        self._framebuffer = None
        self._texture = None
        super().release()

    def create(self, engine, node, resized, **kwargs):
        super().create(engine, node, resized, **kwargs)
        has_alpha = node.get('keep_alpha', 1, bool, False)
        if self._framebuffer is None or self._texture is None or resized or has_alpha != self._has_alpha:
            self._last = None
            self._has_alpha = has_alpha
            self._texture = self.glctx.texture((self.width, self.height), 4 if self._has_alpha else 3,
                                               dtype='f1', internal_format=GL_SRGB8_ALPHA8 if self._has_alpha else GL_SRGB8)
            self._framebuffer = self.glctx.framebuffer(color_attachments=(self._texture,))
            self._framebuffer.clear()

    def render(self, node, **kwargs):
        if filename := node.get('filename', 1, str):
            super().render(node, **kwargs)
            path = SharedCache[filename]
            ext = path.suffix.lower()
            codec = node.get('codec', 1, str, 'h264')
            if ext in ('.mp4', '.mov', '.m4v', '.mkv', '.webm', '.ogg') or (ext == '.gif' and codec == 'gif'):
                pixfmt = node.get('pixfmt', 1, str, 'rgb8' if codec == 'gif' else 'yuv420p')
                crf = node.get('crf', 1, int)
                preset = node.get('preset', 1, str)
                limit = node.get('limit', 1, float)
                path.write_video_frame(self._texture, kwargs['clock'],
                                       fps=int(kwargs['fps']), realtime=kwargs['realtime'], codec=codec,
                                       pixfmt=pixfmt, crf=crf, preset=preset, limit=limit)
            else:
                quality = node.get('quality', 1, int)
                path.write_image(self._texture, quality=quality)


SCENE_NODE_CLASS = Record



# cython: language_level=3, profile=False

from ..model cimport Vector


cdef class uniform(Vector):
    cdef double _item(self, unsigned long long i) noexcept nogil
    cpdef Vector slice(self, Vector index)
    cpdef bint as_bool(self)


cdef class beta(uniform):
    pass


cdef class normal(uniform):
    cdef bint cached
    cdef unsigned long long i
    cdef double R
    cdef double th


cpdef shuffle(uniform source, Vector xs)



"""
Tests of the model.Matrix33 class
"""

import math
import unittest

from flitter.model import Matrix33


def all_isclose(xs, ys, rel_tol=1e-9, abs_tol=0):
    for x, y in zip(xs, ys):
        if not math.isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol):
            return False
    return True


class TestMatrix33(unittest.TestCase):
    """
    Tests of the Matrix33 class

    Note that matrices are column-major as per OpenGL.
    """

    def test_construct(self):
        self.assertEqual(Matrix33(), [1, 0, 0, 0, 1, 0, 0, 0, 1])
        self.assertEqual(Matrix33(0), [0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(Matrix33(1), [1, 0, 0, 0, 1, 0, 0, 0, 1])
        self.assertEqual(Matrix33(2), [2, 0, 0, 0, 2, 0, 0, 0, 2])
        self.assertEqual(Matrix33(range(9)), range(9))
        self.assertRaises(ValueError, Matrix33, "Hello world!")
        self.assertRaises(ValueError, Matrix33, [1, 2, 3])

    def test_translate(self):
        self.assertIsNone(Matrix33.translate(None))
        self.assertIsNone(Matrix33.translate([0, 1, 2]))
        self.assertIsNone(Matrix33.translate("Hello world!"))
        self.assertEqual(Matrix33.translate([2]), [1, 0, 0, 0, 1, 0, 2, 2, 1])
        self.assertEqual(Matrix33.translate([2, 3]), [1, 0, 0, 0, 1, 0, 2, 3, 1])

    def test_scale(self):
        self.assertIsNone(Matrix33.scale(None))
        self.assertIsNone(Matrix33.scale([0, 1, 2]))
        self.assertIsNone(Matrix33.scale("Hello world!"))
        self.assertEqual(Matrix33.scale(2), [2, 0, 0, 0, 2, 0, 0, 0, 1])
        self.assertEqual(Matrix33.scale([2, 3]), [2, 0, 0, 0, 3, 0, 0, 0, 1])

    def test_rotate(self):
        self.assertIsNone(Matrix33.rotate(math.nan))
        self.assertEqual(Matrix33.rotate(0), Matrix33())
        self.assertTrue(all_isclose(Matrix33.rotate(1), Matrix33(), abs_tol=1e-12))
        self.assertTrue(all_isclose(Matrix33.rotate(0.5), [-1, 0, 0, 0, -1, 0, 0, 0, 1], abs_tol=1e-12))

    def test_mmul(self):
        self.assertEqual(Matrix33() @ Matrix33(), Matrix33())
        self.assertEqual(Matrix33.translate([1, 2]) @ Matrix33(), Matrix33.translate([1, 2]))
        self.assertEqual(Matrix33() @ Matrix33.translate([1, 2]), Matrix33.translate([1, 2]))
        self.assertEqual(Matrix33.translate([1, 2]) @ Matrix33.translate([1, 2]), Matrix33.translate([2, 4]))
        self.assertEqual(Matrix33.translate([1, 2]) @ Matrix33.scale(2), [2, 0, 0, 0, 2, 0, 1, 2, 1])
        self.assertEqual(Matrix33.scale(2) @ Matrix33.translate([1, 2]), [2, 0, 0, 0, 2, 0, 2, 4, 1])

    def test_vmul(self):
        self.assertEqual(Matrix33() @ None, None)
        self.assertEqual(Matrix33() @ [], None)
        self.assertEqual(Matrix33() @ [1], None)
        self.assertEqual(Matrix33() @ [1, 2], [1, 2])
        self.assertEqual(Matrix33() @ [1, 2, 3], [1, 2, 3])
        self.assertEqual(Matrix33() @ [1, 2, 3, 1], None)
        self.assertEqual(Matrix33.translate([1, 2]) @ [0, 0], [1, 2])
        self.assertEqual(Matrix33.translate([1, 2]) @ [-1, -2, 1], [0, 0, 1])
        self.assertEqual(Matrix33.scale(2) @ [-1, -2, 1], [-2, -4, 1])

    def test_inverse(self):
        a = Matrix33.scale([2, 3])
        b = Matrix33.rotate(0.25)
        c = Matrix33.translate([7, 9])
        self.assertTrue(all_isclose(a.inverse() @ a, Matrix33()))
        self.assertTrue(all_isclose(b.inverse() @ b, Matrix33()))
        self.assertTrue(all_isclose(b.inverse() @ (a.inverse() @ (a @ (b @ c))), c, abs_tol=1e-12))
        self.assertTrue(all_isclose((((b.inverse() @ a.inverse()) @ a) @ b) @ c, c))

    def test_transpose(self):
        self.assertEqual(Matrix33().transpose(), Matrix33())
        self.assertEqual(Matrix33(range(9)).transpose(), [0, 3, 6, 1, 4, 7, 2, 5, 8])

    def test_repr(self):
        self.assertEqual(repr(Matrix33()), """|   1.000   0.000   0.000 |
|   0.000   1.000   0.000 |
|   0.000   0.000   1.000 |""")



# cython: language_level=3, profile=True

"""
Flitter drawing canvas based on Skia
"""

import array
import functools

import cython
from cpython cimport array, PyObject
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from libc.math cimport acos, sqrt
from libc.stdint cimport int64_t
from loguru import logger
import skia

from . import SceneNode, COLOR_FORMATS
from flitter import name_patch
from ...cache import SharedCache
from ...model cimport Vector, Node
from .glconstants import GL_TEXTURE_2D, GL_RGBA8, GL_RGBA16F, GL_RGBA32F, GL_SRGB8_ALPHA8

cdef extern from "Python.h":
    ctypedef int64_t _PyTime_t
    _PyTime_t _PyTime_GetPerfCounter() noexcept nogil
    double _PyTime_AsSecondsDouble(_PyTime_t t) noexcept nogil


cdef inline double perf_counter() noexcept nogil:
    return _PyTime_AsSecondsDouble(_PyTime_GetPerfCounter())


logger = name_patch(logger, __name__)

cdef double Tau = 6.283185307179586
cdef str DefaultFontFamily = 'Helvetica'
cdef int DefaultFontSize = 20

cdef dict TextureFormatColorType = {
    'f1': (GL_RGBA8, skia.kRGBA_8888_ColorType),
    'f2': (GL_RGBA16F, skia.kRGBA_F16_ColorType),
    'f4': (GL_RGBA32F, skia.kRGBA_F32_ColorType)
}

cdef dict Composite = {
    'clear': skia.BlendMode.kClear,
    'source': skia.BlendMode.kSrc,
    'dest': skia.BlendMode.kDst,
    'over': skia.BlendMode.kSrcOver,
    'dest_over': skia.BlendMode.kDstOver,
    'in': skia.BlendMode.kSrcIn,
    'dest_in': skia.BlendMode.kDstIn,
    'out': skia.BlendMode.kSrcOut,
    'dest_out': skia.BlendMode.kDstOut,
    'atop': skia.BlendMode.kSrcATop,
    'dest_atop': skia.BlendMode.kDstATop,
    'xor': skia.BlendMode.kXor,
    'add': skia.BlendMode.kPlus,
    'modulate': skia.BlendMode.kModulate,
    'screen': skia.BlendMode.kScreen,
    'overlay': skia.BlendMode.kOverlay,
    'darken': skia.BlendMode.kDarken,
    'lighten': skia.BlendMode.kLighten,
    'color_dodge': skia.BlendMode.kColorDodge,
    'color_burn': skia.BlendMode.kColorBurn,
    'hard_light': skia.BlendMode.kHardLight,
    'soft_light': skia.BlendMode.kSoftLight,
    'difference': skia.BlendMode.kDifference,
    'exclusion': skia.BlendMode.kExclusion,
    'multiply': skia.BlendMode.kMultiply,
    'hue': skia.BlendMode.kHue,
    'saturation': skia.BlendMode.kSaturation,
    'color': skia.BlendMode.kColor,
    'luminosity': skia.BlendMode.kLuminosity,
}

cdef dict StrokeJoin = {
    'miter': skia.Paint.Join.kMiter_Join,
    'round': skia.Paint.Join.kRound_Join,
    'bevel': skia.Paint.Join.kBevel_Join,
}

cdef dict StrokeCap = {
    'butt': skia.Paint.Cap.kButt_Cap,
    'round': skia.Paint.Cap.kRound_Cap,
    'square': skia.Paint.Cap.kSquare_Cap,
}

cdef dict FontWeight = {
    'black': skia.FontStyle.Weight.kBlack_Weight,
    'bold': skia.FontStyle.Weight.kBold_Weight,
    'extra_black': skia.FontStyle.Weight.kExtraBlack_Weight,
    'extra_bold': skia.FontStyle.Weight.kExtraBold_Weight,
    'extra_light': skia.FontStyle.Weight.kExtraLight_Weight,
    'invisible': skia.FontStyle.Weight.kInvisible_Weight,
    'light': skia.FontStyle.Weight.kLight_Weight,
    'medium': skia.FontStyle.Weight.kMedium_Weight,
    'normal': skia.FontStyle.Weight.kNormal_Weight,
    'semi_bold': skia.FontStyle.Weight.kSemiBold_Weight,
    'thin': skia.FontStyle.Weight.kThin_Weight,
}

cdef dict FontWidth = {
    'condensed': skia.FontStyle.Width.kCondensed_Width,
    'expanded': skia.FontStyle.Width.kExpanded_Width,
    'extra_condensed': skia.FontStyle.Width.kExtraCondensed_Width,
    'extra_expanded': skia.FontStyle.Width.kExtraExpanded_Width,
    'normal': skia.FontStyle.Width.kNormal_Width,
    'semi_condensed': skia.FontStyle.Width.kSemiCondensed_Width,
    'semi_expanded': skia.FontStyle.Width.kSemiExpanded_Width,
    'ultra_condensed': skia.FontStyle.Width.kUltraCondensed_Width,
    'ultra_expanded': skia.FontStyle.Width.kUltraExpanded_Width,
}

cdef dict FontSlant = {
    'italic': skia.FontStyle.Slant.kItalic_Slant,
    'oblique': skia.FontStyle.Slant.kOblique_Slant,
    'upright': skia.FontStyle.Slant.kUpright_Slant,
}

cdef dict FillType = {
    'even_odd': skia.PathFillType.kEvenOdd,
    'inverse_even_odd': skia.PathFillType.kInverseEvenOdd,
    'inverse_winding': skia.PathFillType.kInverseWinding,
    'winding': skia.PathFillType.kWinding,
}

cdef FillStyle = skia.Paint.Style.kFill_Style
cdef StrokeStyle = skia.Paint.Style.kStroke_Style
cdef StrokeAndFillStyle = skia.Paint.Style.kStrokeAndFill_Style


@cython.cdivision(True)
cdef double turn_angle(double x0, double y0, double x1, double y1, double x2, double y2) noexcept:
    cdef double xa=x1-x0, ya=y1-y0, xb=x2-x1, yb=y2-y1
    cdef double la=sqrt(xa*xa + ya*ya), lb=sqrt(xb*xb + yb*yb)
    if la == 0 or lb == 0:
        return 0
    return acos(min(max(0, (xa*xb + ya*yb) / (la*lb)), 1)) / Tau


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef object line_path(object path, Vector points, double curve, bint closed):
    cdef bint sharp = curve <= 0
    cdef int nverbs = 0, npoints = 0, n = points.length // 2
    if n == 0:
        return
    cdef array.array points_array = array.array('f')
    array.resize(points_array, n * 4 + 4)
    cdef float[:] ppoints = points_array
    cdef array.array verbs_array = array.array('B')
    array.resize(verbs_array, n * 2 + 2)
    cdef unsigned char[:] pverbs = verbs_array
    cdef double last_mid_x, last_mid_y, last_x, last_y, x, y
    cdef int i = 0, m = n, j
    if closed and not sharp:
        m += 2
    while i < m:
        j = i % n * 2
        x, y = points.numbers[j], points.numbers[j + 1]
        if i == 0:
            if sharp or not closed:
                # move to x,y
                pverbs[nverbs] = 0
                nverbs += 1
                ppoints[npoints] = x
                ppoints[npoints+1] = y
                npoints += 2
        elif sharp:
            # line to x,y
            pverbs[nverbs] = 1
            nverbs += 1
            ppoints[npoints] = x
            ppoints[npoints+1] = y
            npoints += 2
        else:
            mid_x, mid_y = (last_x + x) / 2, (last_y + y) / 2
            if i == 1:
                if closed:
                    # move to mid_x,mid_y
                    pverbs[nverbs] = 0
                    nverbs += 1
                    ppoints[npoints] = mid_x
                    ppoints[npoints+1] = mid_y
                    npoints += 2
                else:
                    # line to mid_x,mid_y
                    pverbs[nverbs] = 1
                    nverbs += 1
                    ppoints[npoints] = mid_x
                    ppoints[npoints+1] = mid_y
                    npoints += 2
            elif curve >= 0.5 or turn_angle(last_mid_x, last_mid_y, last_x, last_y, mid_x, mid_y) <= curve:
                # curve to mid_x,mid_y via last_x,last_y
                pverbs[nverbs] = 2
                nverbs += 1
                ppoints[npoints] = last_x
                ppoints[npoints+1] = last_y
                npoints += 2
                ppoints[npoints] = mid_x
                ppoints[npoints+1] = mid_y
                npoints += 2
            else:
                # line to last_x,last_y
                pverbs[nverbs] = 1
                nverbs += 1
                ppoints[npoints] = last_x
                ppoints[npoints+1] = last_y
                npoints += 2
                # line to mid_x,mid_y
                pverbs[nverbs] = 1
                nverbs += 1
                ppoints[npoints] = mid_x
                ppoints[npoints+1] = mid_y
                npoints += 2
            if not closed and i == m-1:
                # line to x,y
                pverbs[nverbs] = 1
                nverbs += 1
                ppoints[npoints] = x
                ppoints[npoints+1] = y
                npoints += 2
            last_mid_x, last_mid_y = mid_x, mid_y
        last_x, last_y = x, y
        i += 1
    if closed:
        # close
        pverbs[nverbs] = 5
        nverbs += 1
    cdef bytearray data = bytearray()
    cdef array.array header_array = array.array('I')
    array.resize(header_array, 4)
    cdef unsigned int[:] header = header_array
    header[0] = 5
    header[1] = npoints / 2
    header[2] = 0
    header[3] = nverbs
    data.extend(header_array)
    array.resize(points_array, npoints)
    data.extend(points_array)
    if nverbs % 4 != 0:
        nverbs += 4 - nverbs % 4
    array.resize(verbs_array, nverbs)
    data.extend(verbs_array)
    cdef int processed
    if path.isEmpty():
        processed = path.readFromMemory(data)
    else:
        new_path = skia.Path()
        processed = new_path.readFromMemory(data)
        path.addPath(new_path)
    if processed != len(data):
        raise ValueError("Bad path data")


cdef object get_color(Node node, default=None):
    cdef Vector v = node._attributes.get('color') if node._attributes else None
    cdef double r, g, b, a
    if v is not None and v.numbers != NULL:
        if v.length == 1:
            r = v.numbers[0]
            return skia.Color4f(r, r, r, 1.)
        if v.length == 3:
            r = v.numbers[0]
            g = v.numbers[1]
            b = v.numbers[2]
            return skia.Color4f(r, g, b, 1.)
        if v.length == 4:
            r = v.numbers[0]
            g = v.numbers[1]
            b = v.numbers[2]
            a = v.numbers[3]
            return skia.Color4f(r, g, b, a)
    return default


cdef void update_context(Node node, ctx):
    cdef str key
    cdef Vector value
    if not node._attributes:
        return
    for key, value in node._attributes.items():
        if key == 'translate':
            if (translate := value.match(2, float)) is not None:
                ctx.translate(*translate)
        elif key == 'rotate':
            if (rotate := value.match(1, float)) is not None:
                ctx.rotate(rotate * 360)
        elif key == 'scale':
            if (scale := value.match(2, float)) is not None:
                ctx.scale(*scale)


cdef void update_path(Node node, path):
    cdef str key
    cdef Vector value
    if not node._attributes:
        return
    for key, value in node._attributes.items():
        if key == 'fill_type':
            if (fill_type := FillType.get(value.match(1, str))) is not None:
                path.setFillType(fill_type)


cdef object update_paint(Node node, start_paint, colorspace):
    cdef str key
    cdef Vector value
    if not node._attributes:
        return start_paint
    paint = start_paint
    for key, value in node._attributes.items():
        if key == 'color':
            if (color := get_color(node)) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setColor4f(color, colorspace)
                if paint.getShader():
                    if colorspace is not None:
                        paint.setShader(skia.Shaders.Color(color, colorspace))
                    else:
                        paint.setShader(skia.Shaders.Color(color.toColor()))
        elif key == 'stroke_width':
            if (stroke_width := value.match(1, float)) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setStrokeWidth(stroke_width)
        elif key == 'stroke_join':
            if (stroke_join := StrokeJoin.get(value.match(1, str))) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setStrokeJoin(stroke_join)
        elif key == 'stroke_cap':
            if (stroke_cap := StrokeCap.get(value.match(1, str))) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setStrokeCap(stroke_cap)
        elif key == 'composite':
            if (composite := Composite.get(value.match(1, str))) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setBlendMode(composite)
        elif key == 'antialias':
            if (antialias := value.match(1, bool)) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setAntiAlias(antialias)
        elif key == 'dither':
            if (dither := value.match(1, bool)) is not None:
                if paint is start_paint:
                    paint = skia.Paint(paint)
                paint.setDither(dither)
    return paint


cdef object update_font(Node node, font):
    if node._attributes is not None and ('font_size' in node._attributes or
                                         'font_family' in node._attributes or 'font_weight' in node._attributes or
                                         'font_width' in node._attributes or 'font_slant' in node._attributes):
        typeface = font.getTypeface()
        font_style = typeface.fontStyle()
        font_family = node.get('font_family', 1, str, typeface.getFamilyName())
        font_weight = FontWeight.get(node.get('font_weight', 1, str))
        weight = skia.FontStyle.Weight(font_weight) if font_weight is not None else font_style.weight()
        font_width = FontWidth.get(node.get('font_width', 1, str))
        width = skia.FontStyle.Width(font_width) if font_width is not None else font_style.width()
        font_slant = FontSlant.get(node.get('font_slant', 1, str))
        slant = skia.FontStyle.Slant(font_slant) if font_slant is not None else font_style.slant()
        font_size = node.get('font_size', 1, float, font.getSize())
        return skia.Font(skia.Typeface(font_family, skia.FontStyle(weight, width, slant)), font_size)
    return font


cdef object make_image(ctx, Node node, dict references):
    if (filename := node.get('filename', 1, str)):
        return SharedCache[filename].read_skia_image()
    elif references is not None and (id := node.get('texture_id', 1, str)) and id in references:
        scene_node = references[id]
        if hasattr(scene_node, 'texture') and scene_node.texture is not None:
            texture = scene_node.texture
            format, colortype = TextureFormatColorType[texture.dtype]
            colorspace = skia.ColorSpace.MakeSRGBLinear()
            texture_info = skia.GrGLTextureInfo(GL_TEXTURE_2D, texture.glo, format)
            backend_texture = skia.GrBackendTexture(texture.width, texture.height, skia.GrMipmapped.kNo, texture_info)
            return skia.Image.MakeFromTexture(ctx.getSurface().recordingContext(), backend_texture,
                                              skia.GrSurfaceOrigin.kBottomLeft_GrSurfaceOrigin,
                                              colortype, skia.AlphaType.kPremul_AlphaType, colorspace)


cdef object make_shader(ctx, Node node, paint, dict references, colorspace):
    cdef list colors, positions, shaders = []
    cdef int i, nstops
    cdef str noise_type, key

    cdef Node child
    for child in node._children:
        shader = make_shader(ctx, child, paint, references, colorspace)
        if shader is not None:
            shaders.append(shader)

    cdef str kind = node.kind
    if kind == 'color':
        color = get_color(node, paint.getColor())
        return skia.Shaders.Color(color, colorspace)

    elif kind == 'gradient':
        colors = []
        positions = []
        for child in node._children:
            if child.kind == 'stop':
                positions.append(child.get('offset', 1, float))
                colors.append(get_color(child, paint.getColor4f()).toColor())
        nstops = len(positions)
        if nstops:
            for i in range(nstops):
                if positions[i] is None:
                    positions[i] = i / (nstops - 1)
            if (start := node.get('start', 2, float)) is not None and (end := node.get('end', 2, float)) is not None:
                points = [skia.Point(*start), skia.Point(*end)]
                return skia.GradientShader.MakeLinear(points, colors, positions)
            if (radius := node.get('radius', 2, float)) is not None:
                rotate = node.get('rotate', 1, float, 0)
                matrix = skia.Matrix.Scale(*radius).postRotate(rotate * 360).postTranslate(*node.get('point', 2, float, (0, 0)))
                return skia.GradientShader.MakeRadial(skia.Point(0, 0), 1, colors, positions, localMatrix=matrix)

    elif kind == 'noise':
        if (frequency := node.get('frequency', 2, float)) is not None:
            octaves = node.get('octaves', 1, int, 8)
            seed = node.get('seed', 1, float, 0)
            size = skia.ISize(*node.get('size', 2, int, (0, 0)))
            noise_type = node.get('type', 1, str, 'improved')
            if noise_type == 'fractal':
                return skia.PerlinNoiseShader.MakeFractalNoise(*frequency, octaves, seed, size)
            elif noise_type == 'turbulence':
                return skia.PerlinNoiseShader.MakeTurbulence(*frequency, octaves, seed, size)
            else:
                return skia.PerlinNoiseShader.MakeImprovedNoise(*frequency, octaves, seed)

    elif kind == 'blend':
        if len(shaders) == 2:
            if (ratio := node.get('ratio', 1, float)) is not None:
                return skia.Shaders.Lerp(min(max(0, ratio), 1), *shaders)
            if (mode := Composite.get(node.get('mode', 1, str, 'over'))) is not None:
                return skia.Shaders.Blend(skia.BlendMode(mode), *shaders)

    elif kind == 'pattern':
        if (image := make_image(ctx, node, references)) is not None:
            matrix = skia.Matrix()
            if node._attributes:
                for key, value in node._attributes.items():
                    if key == 'translate':
                        if (translate := value.match(2, float)) is not None:
                            matrix.postTranslate(*translate)
                    elif key == 'rotate':
                        if (rotate := value.match(1, float)) is not None:
                            matrix.postRotate(rotate * 360)
                    elif key == 'scale':
                        if (scale := value.match(2, float)) is not None:
                            matrix.postScale(*scale)
            return image.makeShader(skia.TileMode.kRepeat, skia.TileMode.kRepeat, localMatrix=matrix)

    return None


cdef object make_image_filter(Node node, paint, colorspace):
    cdef list sub_filters = []
    cdef Node child
    for child in node._children:
        image_filter = make_image_filter(child, paint, colorspace)
        if image_filter is not False:
            sub_filters.append(image_filter)

    cdef str kind = node.kind
    if kind == 'source':
        return None

    elif kind == 'blend':
        if len(sub_filters) == 2:
            if (coefficients := node.get('coefficients', 4, float)) is not None:
                return skia.ImageFilters.Arithmetic(*coefficients, True, *sub_filters)
            if (ratio := node.get('ratio', 1, float)) is not None:
                return skia.ImageFilters.Arithmetic(0, 1-ratio, ratio, 0, True, *sub_filters)
            if (mode := Composite.get(node.get('mode', 1, str))) is not None:
                return skia.ImageFilters.Xfermode(mode, *sub_filters)

    elif kind == 'blur':
        if len(sub_filters) <= 1 and (radius := node.get('radius', 2, float)) is not None:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            return skia.ImageFilters.Blur(*radius, skia.TileMode.kClamp, input=input_filter)

    elif kind == 'shadow':
        if len(sub_filters) <= 1 and (radius := node.get('radius', 2, float)) is not None:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            offset = node.get('offset', 2, float, (0, 0))
            color = get_color(node, paint.getColor())
            if node.get('shadow_only', 1, bool, False):
                return skia.ImageFilters.DropShadowOnly(*offset, *radius, color.toColor(), input=input_filter)
            return skia.ImageFilters.DropShadow(*offset, *radius, color.toColor(), input=input_filter)

    elif kind == 'offset':
        if len(sub_filters) <= 1 and (offset := node.get('offset', 2, float)) is not None:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            return skia.ImageFilters.Offset(*offset, input=input_filter)

    elif kind == 'dilate':
        if len(sub_filters) <= 1 and (radius := node.get('radius', 2, float)) is not None:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            return skia.ImageFilters.Dilate(*radius, input=input_filter)

    elif kind == 'erode':
        if len(sub_filters) <= 1 and (radius := node.get('radius', 2, float)) is not None:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            return skia.ImageFilters.Erode(*radius, input=input_filter)

    elif kind == 'paint':
        if len(sub_filters) == 0:
            paint = update_paint(node, paint, colorspace)
            return skia.ImageFilters.Paint(paint)

    elif kind == 'color_matrix':
        if len(sub_filters) <= 1:
            input_filter = sub_filters[0] if len(sub_filters) == 1 else None
            matrix = node.get('matrix', 20, float)
            if matrix is None:
                if node.keys() & {'red', 'green', 'blue', 'alpha'}:
                    red = node.get('red', 5, float, [1, 0, 0, 0, 0])
                    green = node.get('green', 5, float, [0, 1, 0, 0, 0])
                    blue = node.get('blue', 5, float, [0, 0, 1, 0, 0])
                    alpha = node.get('alpha', 5, float, [0, 0, 0, 1, 0])
                    matrix = red + green + blue + alpha
                elif node.keys() & {'scale', 'offset'}:
                    scale = node.get('scale', 3, float, [1, 1, 1])
                    offset = node.get('offset', 3, float, [0, 0, 0])
                    matrix = [scale[0], 0, 0, 0, offset[0],
                              0, scale[1], 0, 0, offset[1],
                              0, 0, scale[2], 0, offset[2],
                              0, 0, 0, 1, 0]
                elif node.keys() & {'brightness', 'contrast', 'saturation'}:
                    brightness = node.get('brightness', 1, float, 0)
                    contrast = node.get('contrast', 1, float, 1)
                    saturation = node.get('saturation', 1, float, 1)
                    offset = brightness + (1 - contrast) / 2
                    s0 = contrast * (1 - saturation)
                    s1 = contrast * saturation
                    rs0, gs0, bs0 = s0*0.2126, s0*0.7152, s0*0.0722
                    matrix = [rs0+s1, gs0, bs0, 0, offset,
                              rs0, gs0+s1, bs0, 0, offset,
                              rs0, gs0, bs0+s1, 0, offset,
                              0, 0, 0, 1, 0]
            if matrix is not None:
                color_filter = skia.ColorFilters.Matrix(matrix)
                return skia.ImageFilters.ColorFilter(color_filter, input=input_filter)

    return False


cdef object make_path_effect(Node node):
    cdef list sub_path_effects = []
    cdef Node child
    for child in node._children:
        path_effect = make_path_effect(child)
        if path_effect is not False:
            sub_path_effects.append(path_effect)

    cdef str kind = node.kind
    if kind == 'dash':
        if len(sub_path_effects) <= 1 and (intervals := node.get('intervals', 0, float)):
            offset = node.get('offset', 1, float, 0)
            path_effect = skia.DashPathEffect.Make(intervals, offset)
            return path_effect if not sub_path_effects else skia.PathEffect.MakeCompose(path_effect, sub_path_effects[0])

    elif kind == 'round_corners':
        if len(sub_path_effects) <= 1 and (radius := node.get('radius', 1, float)):
            path_effect = skia.CornerPathEffect.Make(radius)
            return path_effect if not sub_path_effects else skia.PathEffect.MakeCompose(path_effect, sub_path_effects[0])

    elif kind == 'jitter':
        if len(sub_path_effects) <= 1 and (length := node.get('length', 1, float)) and (deviation := node.get('deviation', 1, float)):
            seed = node.get('seed', 1, int, 0)
            path_effect = skia.DiscretePathEffect.Make(length, deviation, seed)
            return path_effect if not sub_path_effects else skia.PathEffect.MakeCompose(path_effect, sub_path_effects[0])

    elif kind == 'path_matrix':
        if len(sub_path_effects) <= 1:
            matrix = node.get('matrix', 9, float)
            if matrix is None and node.keys() & {'scale', 'rotate', 'translate'}:
                scale = node.get('scale', 2, float)
                matrix = skia.Matrix.Scale(*scale) if scale else skia.Matrix.I()
                rotate = node.get('rotate', 1, float)
                if rotate:
                    matrix = skia.Matrix.Concat(skia.Matrix.RotateDeg(rotate * 360), matrix)
                translate = node.get('translate', 2, float)
                if translate:
                    matrix = skia.Matrix.Concat(skia.Matrix.Translate(*translate), matrix)
            if matrix is not None:
                path_effect = skia.MatrixPathEffect.Make(skia.Matrix.MakeAll(*matrix))
                return path_effect if not sub_path_effects else skia.PathEffect.MakeCompose(path_effect, sub_path_effects[0])

    elif kind == 'sum':
        if len(sub_path_effects) >= 2:
            return functools.reduce(skia.PathEffect.MakeSum, sub_path_effects)

    return None


cpdef object draw(Node node, ctx, paint=None, font=None, path=None, dict stats=None, dict references=None, colorspace=None):
    cdef double duration, total, start_time = perf_counter()
    cdef int count
    ctx.save()
    update_context(node, ctx)
    paint = update_paint(node, skia.Paint(AntiAlias=True) if paint is None else paint, colorspace)
    font = update_font(node, skia.Font(skia.Typeface(DefaultFontFamily), DefaultFontSize) if font is None else font)
    path = skia.Path()
    cdef Node child
    for child in node._children:
        paint = _draw(child, node, ctx, paint, font, path, stats, references, colorspace)
    ctx.restore()
    if stats is not None:
        duration = perf_counter() - start_time
        count, total = stats.get(node.kind, (0, 0))
        stats[node.kind] = count+1, total+duration


cpdef object _draw(Node node, Node parent, ctx, paint, font, path, dict stats, dict references, colorspace):
    cdef double start_time = perf_counter()
    cdef Vector points
    cdef Node child
    cdef str kind = node.kind
    cdef double x, y, rx, ry, sx, sy
    cdef bint fill, stroke

    if kind == 'group':
        ctx.save()
        path = skia.Path()
        update_path(node, path)
        update_context(node, ctx)
        group_paint = update_paint(node, paint, colorspace)
        font = update_font(node, font)
        for child in node._children:
            group_paint = _draw(child, node, ctx, group_paint, font, path, stats, references, colorspace)
        ctx.restore()

    elif kind == 'transform':
        ctx.save()
        update_context(node, ctx)
        for child in node._children:
            paint = _draw(child, node, ctx, paint, font, path, stats, references, colorspace)
        ctx.restore()

    elif kind == 'paint':
        paint_paint = update_paint(node, paint, colorspace)
        for child in node._children:
            paint_paint = _draw(child, node, ctx, paint_paint, font, path, stats, references, colorspace)

    elif kind == 'font':
        font = update_font(node, font)
        for child in node._children:
            paint = _draw(child, node, ctx, paint, font, path, stats, references, colorspace)

    elif kind == 'path':
        path = skia.Path()
        update_path(node, path)
        for child in node._children:
            paint = _draw(child, node, ctx, paint, font, path, stats, references, colorspace)

    elif kind == 'move_to':
        if (point := node.get('point', 2, float)) is not None:
            path.moveTo(*point)

    elif kind == 'line_to':
        if (point := node.get('point', 2, float)) is not None:
            path.lineTo(*point)

    elif kind == 'curve_to':
        if (point := node.get('point', 2, float)) is not None and (c1 := node.get('c1', 2, float)) is not None:
            if (c2 := node.get('c2', 2, float)) is not None:
                path.cubicTo(*c1, *c2, *point)
            else:
                path.quadTo(*c1, *point)

    elif kind == 'arc_to':
        if (point := node.get('point', 2, float)) is not None and (radius := node.get('radius', 2, float)) is not None:
            x, y = point
            rx, ry = radius
            rotate = node.get('rotation', 1, float, 0)
            large = node.get('large', 1, bool, False)
            ccw = node.get('ccw', 1, bool, False)
            path.arcTo(rx, ry, rotate, large, ccw, x, y)

    elif kind == 'arc':
        if (point := node.get('point', 2, float, (0, 0))) is not None and (radius := node.get('radius', 2, float)) is not None:
            x, y = point
            rx, ry = radius
            start = node.get('start', 1, float, 0)
            if (sweep := node.get('sweep', 1, float)) is None:
                end = node.get('end', 1, float)
                if end is not None:
                    sweep = end - start
            if sweep is not None:
                move_to = node.get('move_to', 1, bool, False)
                path.arcTo(skia.Rect(x-rx, y-ry, x+rx, y+ry), start*360, sweep*360, move_to)

    elif kind == 'rect':
        if (size := node.get('size', 2, float)) is not None:
            sx, sy = size
            x, y = node.get('point', 2, float, (0, 0))
            rx, ry = node.get('radius', 2, float, (0, 0))
            if (fill := node.get_bool('fill', False)) or (stroke := node.get_bool('stroke', False)):
                path = skia.Path()
            if rx or ry:
                path.addRRect(skia.RRect(skia.Rect.MakeXYWH(x, y, sx, sy), rx, ry))
            else:
                path.addRect(x, y, x+sx, y+sy)
            if fill or stroke:
                draw_paint = update_paint(node, paint, colorspace)
                draw_paint.setStyle((FillStyle if not stroke else StrokeAndFillStyle) if fill else StrokeStyle)
                ctx.drawPath(path, draw_paint)

    elif kind == 'ellipse':
        if (radius := node.get('radius', 2, float)) is not None:
            rx, ry = radius
            x, y = node.get('point', 2, float, (0, 0))
            if (fill := node.get_bool('fill', False)) or (stroke := node.get_bool('stroke', False)):
                path = skia.Path()
            path.addOval(skia.Rect(x-rx, y-ry, x+rx, y+ry))
            if fill or stroke:
                draw_paint = update_paint(node, paint, colorspace)
                draw_paint.setStyle((FillStyle if not stroke else StrokeAndFillStyle) if fill else StrokeStyle)
                ctx.drawPath(path, draw_paint)

    elif kind == 'line':
        if node._attributes and (points := node._attributes.get('points')) is not None and points.numbers is not NULL:
            curve = node.get('curve', 1, float, 0)
            close = node.get('close', 1, bool, False)
            if (fill := node.get_bool('fill', False)) or (stroke := node.get_bool('stroke', False)):
                path = skia.Path()
            line_path(path, points, curve, close)
            if fill or stroke:
                draw_paint = update_paint(node, paint, colorspace)
                draw_paint.setStyle((FillStyle if not stroke else StrokeAndFillStyle) if fill else StrokeStyle)
                ctx.drawPath(path, draw_paint)

    elif kind == 'close':
        path.close()

    elif kind == 'clip':
        ctx.clipPath(path, skia.ClipOp.kIntersect, paint.isAntiAlias())

    elif kind == 'mask':
        ctx.clipPath(path, skia.ClipOp.kDifference, paint.isAntiAlias())

    elif kind == 'fill':
        fill_paint = update_paint(node, paint, colorspace)
        fill_paint.setStyle(FillStyle)
        ctx.drawPath(path, fill_paint)

    elif kind == 'stroke':
        stroke_paint = update_paint(node, paint, colorspace)
        stroke_paint.setStyle(StrokeStyle)
        ctx.drawPath(path, stroke_paint)

    elif kind == 'text':
        if (text := node.get('text', 1, str)) is not None:
            x, y = node.get('point', 2, float, (0, 0))
            text_paint = update_paint(node, paint, colorspace)
            font = update_font(node, font)
            stroke = node.get('stroke', 1, bool, False)
            fill = node.get('fill', 1, bool, not stroke)
            text_paint.setStyle((FillStyle if not stroke else StrokeAndFillStyle) if fill else StrokeStyle)
            if node.get('center', 1, bool, True):
                bounds = skia.Rect(0, 0, 0, 0)
                font.measureText(text, bounds=bounds, paint=text_paint)
                rx = (bounds.fLeft + bounds.fRight)/2
                ry = (bounds.fBottom + bounds.fTop)/2
                ctx.drawString(text, x-rx, y-ry, font, text_paint)
            else:
                ctx.drawString(text, x, y, font, text_paint)

    elif kind == 'image':
        if (image := make_image(ctx, node, references)) is not None:
            width, height = image.width(), image.height()
            point = node.get('point', 2, float, (0, 0))
            if (fill_size := node.get('fill', 2, float)) is not None:
                aspect = fill_size[0] / fill_size[1]
                if width/height > aspect:
                    w = height * aspect
                    src = skia.Rect.MakeXYWH((width-w)/2, 0, w, height)
                else:
                    h = width / aspect
                    src = skia.Rect.MakeXYWH(0, (height-h)/2, width, h)
                dst = skia.Rect.MakeXYWH(*point, *fill_size)
            elif (fit := node.get('fit', 2, float)) is not None:
                aspect = width / height
                src = skia.Rect.MakeXYWH(0, 0, width, height)
                x, y = point
                if fit[0]/fit[1] > aspect:
                    w = fit[1] * aspect
                    dst = skia.Rect.MakeXYWH(x+(fit[0]-w)/2, y, w, fit[1])
                else:
                    h = fit[0] / aspect
                    dst = skia.Rect.MakeXYWH(x, y+(fit[1]-h)/2, fit[0], h)
            else:
                size = node.get('size', 2, float, (width, height))
                src = skia.Rect.MakeXYWH(0, 0, width, height)
                dst = skia.Rect.MakeXYWH(*point, *size)
            ctx.drawImageRect(image, src, dst, paint=paint)

    elif kind == 'layer':
        if (size := node.get('size', 2, float)) is not None:
            alpha = node.get('alpha', 1, float, 1)
            origin = node.get('origin', 2, float, (0, 0))
            rect = skia.Rect.MakeXYWH(*origin, *size)
            ctx.clipRect(rect)
            ctx.saveLayerAlpha(rect, int(alpha * 255))
            path = skia.Path()
            update_context(node, ctx)
            layer_paint = update_paint(node, paint, colorspace)
            font = update_font(node, font)
            for child in node._children:
                layer_paint = _draw(child, node, ctx, layer_paint, font, path, stats, references, colorspace)
            ctx.restore()

    elif shader := make_shader(ctx, node, paint, references, colorspace):
        paint = skia.Paint(paint)
        paint.setShader(shader)

    elif image_filter := make_image_filter(node, paint, colorspace):
        paint = skia.Paint(paint)
        paint.setImageFilter(image_filter)

    elif path_effect := make_path_effect(node):
        paint = skia.Paint(paint)
        paint.setPathEffect(path_effect)

    cdef double duration, total
    cdef int count
    cdef str parent_kind
    cdef PyObject* objptr
    if stats is not None:
        duration = perf_counter() - start_time
        objptr = PyDict_GetItem(stats, kind)
        if objptr != NULL:
            count, total = <tuple>objptr
            PyDict_SetItem(stats, kind, (count+1, total+duration))
        else:
            PyDict_SetItem(stats, kind, (1, duration))
        parent_kind = parent.kind
        objptr = PyDict_GetItem(stats, parent_kind)
        if objptr != NULL:
            count, total = <tuple>objptr
            PyDict_SetItem(stats, parent_kind, (count, total-duration))
        else:
            PyDict_SetItem(stats, parent_kind, (0, -duration))

    return paint


class Canvas(SceneNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._graphics_context = skia.GrDirectContext.MakeGL()
        self._texture = None
        self._framebuffer = None
        self._surface = None
        self._canvas = None
        self._stats = {}
        self._total_duration = 0
        self._colorbits = None
        self._linear = None
        self._colorspace = None

    @property
    def texture(self):
        return self._texture

    def release(self):
        self._colorbits = None
        self._linear = None
        self._colorspace = None
        self._canvas = None
        self._surface = None
        if self._graphics_context is not None:
            self._graphics_context.abandonContext()
            self._graphics_context = None
        self._framebuffer = None
        self._texture = None

    def create(self, engine, node, resized, **kwargs):
        colorbits = node.get('colorbits', 1, int, self.glctx.extra['colorbits'])
        if colorbits not in COLOR_FORMATS:
            colorbits = self.glctx.extra['colorbits']
        linear = node.get('linear', 1, bool, self.glctx.extra['linear'])
        if resized or colorbits != self._colorbits or linear != self._linear:
            depth = COLOR_FORMATS[colorbits]
            skia_colortype = TextureFormatColorType[depth.moderngl_dtype][1]
            internal_format = None if linear else GL_SRGB8_ALPHA8
            self._colorspace = skia.ColorSpace.MakeSRGBLinear() if linear else skia.ColorSpace.MakeSRGB()
            self._texture = self.glctx.texture((self.width, self.height), 4, dtype=depth.moderngl_dtype, internal_format=internal_format)
            self._framebuffer = self.glctx.framebuffer(color_attachments=(self._texture,))
            backend_render_target = skia.GrBackendRenderTarget(self.width, self.height, 0, 0, skia.GrGLFramebufferInfo(self._framebuffer.glo, depth.gl_format))
            self._surface = skia.Surface.MakeFromBackendRenderTarget(self._graphics_context, backend_render_target, skia.kBottomLeft_GrSurfaceOrigin,
                                                                     skia_colortype, self._colorspace)
            self._canvas = self._surface.getCanvas()
            self._colorbits = colorbits
            self._linear = linear
            logger.debug("Created {:d}x{:d} canvas; skia version {}", self.width, self.height, skia.__version__)

    async def descend(self, engine, node, **kwargs):
        # A canvas is a leaf node from the perspective of the OpenGL world
        pass

    def purge(self):
        total_count = self._stats['canvas'][0]
        logger.info("{} render stats - {:d} x {:.1f}ms = {:.1f}s", self.name, total_count,
                    1e3 * self._total_duration / total_count, self._total_duration)
        draw_duration = 0
        for duration, count, key in sorted(((duration, count, key) for (key, (count, duration)) in self._stats.items()), reverse=True):
            logger.debug("{:15s}  - {:8d}  x {:6.1f}µs = {:5.1f}s  ({:4.1f}%)",
                         key, count, 1e6 * duration / count, duration, 100 * duration / self._total_duration)
            draw_duration += duration
        overhead = self._total_duration - draw_duration
        logger.debug("{:15s}  - {:8d}  x {:6.1f}µs = {:5.1f}s  ({:4.1f}%)",
                     '(surface)', total_count, 1e6 * overhead / total_count, overhead, 100 * overhead / self._total_duration)
        self._stats = {}
        self._total_duration = 0

    def render(self, node, references=None, **kwargs):
        self._total_duration -= perf_counter()
        self._graphics_context.resetContext()
        self._framebuffer.clear()
        draw(node, self._canvas, stats=self._stats, references=references, colorspace=self._colorspace)
        self._surface.flushAndSubmit()
        self._total_duration += perf_counter()


SCENE_NODE_CLASS = Canvas



"""
Tests of the model.Node class
"""

import unittest

from flitter.model import Node, Vector, null

from test_vector import FOO_SYMBOL_NUMBER


class TestNode(unittest.TestCase):
    """
    Tests of the Node class
    """

    def setUp(self):
        self.node1 = Node('node1')
        self.node2 = Node('node2', {'bar', 'baz'})
        self.node3 = Node('node3', {'bar', 'baz'}, {'color': Vector(1)})
        self.node4 = Node('node4')
        self.node5 = Node('node5')

    def test_construct(self):
        self.assertEqual(self.node1.kind, 'node1')
        self.assertEqual(self.node1.tags, set())
        self.assertEqual(len(self.node1), 0)
        self.assertEqual(list(self.node1.items()), [])
        self.assertEqual(list(self.node1.children), [])
        self.assertEqual(self.node2.kind, 'node2')
        self.assertEqual(self.node2.tags, {'bar', 'baz'})
        self.assertEqual(len(self.node2), 0)
        self.assertEqual(list(self.node2.items()), [])
        self.assertEqual(list(self.node2.children), [])
        self.assertEqual(self.node3.kind, 'node3')
        self.assertEqual(self.node3.tags, {'bar', 'baz'})
        self.assertEqual(len(self.node3), 1)
        self.assertEqual(list(self.node3.items()), [('color', Vector(1))])
        self.assertEqual(list(self.node3.children), [])

    def test_children(self):
        self.node1.append(self.node2)
        self.node1.append(self.node3)
        children = self.node1.children
        self.assertIs(next(children), self.node2)
        self.assertIs(next(children), self.node3)
        self.assertRaises(StopIteration, next, children)

    def test_copy(self):
        self.node2.append(self.node3)
        self.node1.append(self.node2)
        node1c = self.node1.copy()
        self.assertIsNot(node1c, self.node1)
        self.assertEqual(node1c, self.node1)
        node2c, = node1c.children
        node3c, = node2c.children
        self.assertIs(node2c, self.node2)
        self.assertIs(node3c, self.node3)

    def test_add_tag(self):
        self.node1.add_tag('test')
        self.assertEqual(self.node1.tags, {'test'})
        self.node2.add_tag('test')
        self.assertEqual(self.node2.tags, {'bar', 'baz', 'test'})

    def test_append(self):
        self.node2.append(self.node3)
        self.node1.append(self.node2)
        self.node1.append(self.node3)
        children = list(self.node1.children)
        self.assertEqual(len(children), 2)
        self.assertIs(children[0], self.node2)
        self.assertIs(children[1], self.node3)
        children = list(self.node2.children)
        self.assertEqual(len(children), 1)
        self.assertIs(children[0], self.node3)

    def test_attributes(self):
        self.assertTrue('color' in self.node3)
        self.assertEqual(self.node3['color'], Vector(1))
        self.node3.set_attribute('shape', Vector('circle'))
        shape = self.node3['shape']
        self.assertIsInstance(shape, Vector)
        self.assertEqual(shape, Vector('circle'))
        size = Vector([1, 2, 3])
        self.node3.set_attribute('size', size)
        self.assertIs(self.node3['size'], size)
        self.assertEqual(set(self.node3), {'color', 'shape', 'size'})
        self.assertEqual(self.node3.keys(), {'color', 'shape', 'size'})
        self.assertEqual(list(self.node3.values()), [Vector(1), Vector('circle'), size])
        self.assertEqual(list(self.node3.items()), [('color', Vector(1)), ('shape', Vector('circle')), ('size', size)])
        self.node3.set_attribute('size', null)
        self.assertFalse('size' in self.node3)
        with self.assertRaises(KeyError):
            self.node3['size']

    def test_get(self):
        self.node1.set_attribute('str1', Vector('circle'))
        self.node1.set_attribute('str2', Vector(('Hello ', 'world!')))
        self.node1.set_attribute('float1', Vector(1.5))
        self.node1.set_attribute('float2', Vector((0, 0)))
        self.node1.set_attribute('float3', Vector((1, 2, 3.5)))
        self.node1.set_attribute('symbol', Vector.symbol('foo'))
        self.assertEqual(self.node1.get('missing'), None)
        self.assertEqual(self.node1.get('missing', 0, float), None)
        self.assertEqual(self.node1.get('missing', 1, str), None)
        self.assertEqual(self.node1.get('missing', default=5), 5)
        self.assertEqual(self.node1.get('str1'), ['circle'])
        self.assertEqual(self.node1.get('str1', 1), 'circle')
        self.assertEqual(self.node1.get('str1', 1), 'circle')
        self.assertEqual(self.node1.get('str2', 1), None)
        self.assertEqual(self.node1.get('str2', 1, str), 'Hello world!')
        self.assertEqual(self.node1.get('str2', 1, float), None)
        self.assertEqual(self.node1.get('str2', 2, float), None)
        self.assertEqual(self.node1.get('float1', 0, float), [1.5])
        self.assertEqual(self.node1.get('float1', 1, float), 1.5)
        self.assertEqual(self.node1.get('float1', 3, float), [1.5, 1.5, 1.5])
        self.assertEqual(self.node1.get('float3', 0, float), [1, 2, 3.5])
        self.assertEqual(self.node1.get('float3', 1, float), None)
        self.assertEqual(self.node1.get('float3', 3, float), [1, 2, 3.5])
        self.assertEqual(self.node1.get('float3', 3, int), [1, 2, 3])
        self.assertEqual(self.node1.get('str1', 1, bool), True)
        self.assertEqual(self.node1.get('str2', 1, bool), True)
        self.assertEqual(self.node1.get('str2', 2, bool), [True, True])
        self.assertEqual(self.node1.get('str2', 3, bool), None)
        self.assertEqual(self.node1.get('float3', 1, bool), True)
        self.assertEqual(self.node1.get('float2', 1, bool), False)
        self.assertEqual(self.node1.get('float2', 2, bool), [False, False])
        self.assertEqual(self.node1.get('float2', 3, bool), None)
        self.assertEqual(self.node1.get('symbol', 1, str), 'foo')
        self.assertEqual(self.node1.get('symbol', 1, float), FOO_SYMBOL_NUMBER)

    def test_repr(self):
        self.assertEqual(repr(self.node1), '!node1')
        self.assertEqual(repr(self.node2), '!node2 #bar #baz')
        self.assertEqual(repr(self.node3), '!node3 #bar #baz color=1')
        self.node3.append(self.node4)
        self.node2.append(self.node3)
        self.node1.append(self.node2)
        self.node1.append(self.node5)
        self.node1.set_attribute('foo', Vector.symbol('foo'))
        self.assertEqual(repr(self.node1), """
!node1 foo=:foo
 !node2 #bar #baz
  !node3 #bar #baz color=1
   !node4
 !node5
""".strip())



"""
Tests of the model.Vector class
"""

import math
import unittest

from flitter.model import Vector, true, false, null, Node


FOO_SYMBOL_NUMBER = float.fromhex('-0x1.dcb27518fed9dp+1023')


def test_func():
    pass


class test_class:
    def __len__(self):
        return 0


def all_isclose(xs, ys):
    for x, y in zip(xs, ys):
        if not (math.isclose(x, y) or math.isnan(x) and math.isnan(y)):
            return False
    return True


class TestVector(unittest.TestCase):
    """
    Tests of the Vector class
    """

    def test_construct(self):
        """Test constructor and coerce method"""
        for constructor in (Vector, Vector.coerce):
            for value in [None, [], (), {}, set()]:
                with self.subTest(value=value, constructor=constructor):
                    vector = constructor(value)
                    self.assertIsInstance(vector, Vector)
                    self.assertEqual(len(vector), 0)
                    self.assertFalse(vector.numeric)
                    self.assertFalse(vector.non_numeric)
            for value in [3, 3.4, "Three", True, False, -1, test_func, test_class]:
                for container in [lambda xs: xs[0], list, tuple, set, Vector, lambda xs: {x: None for x in xs}]:
                    with self.subTest(value=value, constructor=constructor, container=container):
                        values = container([value])
                        vector = constructor(values)
                        self.assertIsInstance(vector, Vector)
                        self.assertEqual(len(vector), 1)
                        self.assertEqual(vector[0], value)
                        self.assertTrue(vector.numeric == isinstance(value, (float, int)))
                        self.assertTrue(vector.non_numeric == (not vector.numeric))
            for value in [Node('foo'), Node('bar', {'baz'}, {'color': Vector(1)})]:
                for container in [lambda xs: xs[0], list, tuple, Vector]:
                    with self.subTest(value=value, constructor=constructor, container=container):
                        values = container([value])
                        vector = constructor(values)
                        self.assertIsInstance(vector, Vector)
                        self.assertEqual(len(vector), 1)
                        self.assertIs(vector[0], value)
                        self.assertFalse(vector.numeric)
                        self.assertTrue(vector.non_numeric)
            for values in [[0, 0.1, 1.5, -1e99, math.inf], {"Hello", "world!"}, [99, "red", "balloons"],
                           (test_func, test_class), {3: "three"}, {"three": 3},
                           range(10), range(-10), range(0, 10), range(0, 10, 1), range(9, -1, -1),
                           [Node('foo'), Node('bar', {'baz'}, {'color': Vector(1)})]]:
                with self.subTest(values=values, constructor=constructor):
                    vector = constructor(values)
                    self.assertIsInstance(vector, Vector)
                    self.assertEqual(len(vector), len(values))
                    all_numbers = True
                    for x, y in zip(vector, values):
                        self.assertEqual(x, y)
                        all_numbers = all_numbers and isinstance(y, (float, int))
                    self.assertEqual(vector.numeric, len(vector) > 0 and all_numbers)
                    self.assertEqual(vector.non_numeric, len(vector) > 0 and not all_numbers)
                    vector2 = Vector.coerce(vector)
                    self.assertIs(vector2, vector)

    def test_special_coercion_values(self):
        self.assertIs(Vector.coerce(None), null)
        self.assertIs(Vector.coerce([]), null)
        self.assertIs(Vector.coerce(()), null)
        self.assertIs(Vector.coerce({}), null)
        self.assertIs(Vector.coerce(set()), null)
        self.assertIs(Vector.coerce(True), true)
        self.assertIs(Vector.coerce(False), false)
        self.assertIs(Vector.coerce(1), true)
        self.assertIs(Vector.coerce(0), false)
        self.assertIs(Vector.coerce(-1), Vector.coerce(-1))

    def test_symbol(self):
        foo = Vector.symbol('foo')
        foo_n = foo[0]
        self.assertIsInstance(foo_n, float)
        self.assertTrue(foo_n < float.fromhex('-1p1023'))
        self.assertEqual(math.floor(foo_n), foo_n)
        self.assertEqual(math.floor(foo), foo)
        self.assertEqual(foo, Vector.symbol('foo'))
        self.assertNotEqual(foo, Vector.symbol('bar'))
        self.assertIs(str(foo), 'foo')

    def test_copy(self):
        for x in (Vector(["Hello ", "world!"]), Vector.range(10)):
            y = Vector.copy(x)
            self.assertEqual(x, y)
            self.assertFalse(x is y)

    def test_compose(self):
        self.assertEqual(Vector.compose([]), null)
        self.assertEqual(Vector.compose([null]), null)
        self.assertEqual(Vector.compose([null, null]), null)
        self.assertEqual(Vector.compose([true, null]), true)
        self.assertEqual(Vector.compose([true, true]), Vector([1, 1]))
        self.assertEqual(Vector.compose([Vector.range(i*100, (i+1)*100) for i in range(10)]), Vector.range(1000))
        self.assertEqual(Vector.compose([Vector("Hello world!"), Vector([1, 2, 3])]), Vector(["Hello world!", 1, 2, 3]))
        foo = Vector.symbol('foo')
        bar = Vector.symbol('bar')
        self.assertEqual(Vector.compose([foo, bar]), Vector([float(foo), float(bar)]))

    def test_indexing(self):
        numbers = Vector.range(10)
        for i in range(-10, 20):
            self.assertEqual(numbers[i], i % 10)
        objects = Vector([str(i) for i in range(10)])
        for i in range(-10, 20):
            self.assertEqual(objects[i], str(i % 10))
        for i in range(-10, 20):
            self.assertEqual(null[i], null)

    def test_slicing(self):
        numbers = Vector.range(10)
        objects = Vector([str(i) for i in range(10)])
        for i in range(-20, 10):
            for j in range(i+2, 10):
                r = Vector.range(i, j)
                self.assertEqual(numbers[r], r % 10)
                self.assertEqual(objects[r], [str(int(i) % 10) for i in r])
                self.assertEqual(null[r], null)

    def test_range_slice(self):
        TESTS = [
            ((None,), []),
            ((10,), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
            ((9.9,), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
            ((10.3,), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            ((0.5, 10.5), [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5]),
            ((10.5, 0.5), []),
            ((0.5, 10.5, 1.5), [0.5, 2, 3.5, 5, 6.5, 8, 9.5]),
            ((0.5, 10.5, -1.5), []),
            ((10.5, 0.5, -1.5), [10.5, 9, 7.5, 6, 4.5, 3, 1.5]),
            ((10.5, 0.5, -10.5), [10.5]),
        ]
        for s, values in TESTS:
            with self.subTest(slice=s):
                vector = Vector(slice(*s))
                self.assertEqual(list(vector), values)
                self.assertTrue(len(values) == 0 or vector.numeric)
                vector = Vector.range(*s)
                self.assertEqual(list(vector), values)
                self.assertTrue(len(values) == 0 or vector.numeric)

    def test_isinstance(self):
        TESTS = [
            (None, []),
            (float, [0, 0.1, 1.5, -1e99, math.nan, math.inf]),
            (str, ["Hello", "world!"]),
            (Node, [Node('foo'), Node('bar', {'baz'}, {'color': Vector(1)})]),
            (None, [99, "red", "balloons"]),
            (float, Vector.symbol('foo')),
        ]
        for t, values in TESTS:
            with self.subTest(type=t, values=values):
                vector = Vector(values)
                if t is not None:
                    self.assertTrue(vector.isinstance(t))
                for t2, _ in TESTS:
                    if t2 is not None and t2 is not t:
                        self.assertFalse(vector.isinstance(t2))

    def test_as_bool(self):
        self.assertFalse(null)
        self.assertFalse(false)
        self.assertFalse(Vector())
        self.assertFalse(Vector(0))
        self.assertFalse(Vector([0, 0, 0]))
        self.assertFalse(Vector([""]))
        self.assertFalse(Vector(["", "", ""]))
        self.assertFalse(Vector([0, "", 0]))
        self.assertTrue(true)
        self.assertTrue(Vector(1))
        self.assertTrue(Vector([0, 0, 1]))
        self.assertTrue(Vector(["foo"]))
        self.assertTrue(Vector(["", "bar"]))
        self.assertTrue(Vector(Node("foo")))
        self.assertTrue(Vector(test_func))
        self.assertTrue(Vector(test_class))
        self.assertTrue(Vector.symbol('foo'))

    def test_as_double(self):
        for value in [0, 1, 0.1, 1.5, -1e99, math.inf, True, False]:
            self.assertEqual(float(Vector(value)), value)
        self.assertTrue(math.isnan(float(Vector(math.nan))))
        self.assertTrue(math.isnan(float(null)))
        self.assertTrue(math.isnan(float(Vector([1, 2, 3]))))
        self.assertTrue(math.isnan(float(Vector("Hello world!"))))
        foo = Vector.symbol('foo')
        self.assertEqual(float(foo), foo[0])

    def test_as_string(self):
        TESTS = [
            (null, ""),
            (Vector([1, -2, 3.0, 1e99, 1.567e-9]), "1-231e+991.567e-09"),
            (Vector([1 / 3]), "0.333333333"),
            (Vector("Hello world!"), "Hello world!"),
            (Vector(["Hello ", "world!"]), "Hello world!"),
            (Vector(["testing", "testing", 1, 2.2, 3.0]), "testingtesting12.23"),
            (Vector.symbol('foo'), "foo"),
            (Vector.symbol('foo').concat(Vector.symbol('bar')), "foobar"),
        ]
        for vector, string in TESTS:
            self.assertEqual(str(vector), string)

    def test_iter(self):
        self.assertRaises(StopIteration, next, iter(null))
        i = iter(Vector([1, 2, 3]))
        self.assertEqual(next(i), 1)
        self.assertEqual(next(i), 2)
        self.assertEqual(next(i), 3)
        self.assertRaises(StopIteration, next, i)
        i = iter(Vector(["Hello", "world!"]))
        self.assertEqual(next(i), "Hello")
        self.assertEqual(next(i), "world!")
        self.assertRaises(StopIteration, next, i)

    def test_hash(self):
        self.assertEqual(hash(null), 0xe220a8397b1dcdaf - (1 << 64))
        self.assertEqual(hash(Vector()), 0xe220a8397b1dcdaf - (1 << 64))
        self.assertEqual(hash(Vector(0)), -6411193824288604561)
        self.assertEqual(hash(Vector(0.1)), -7435092146473341298)
        self.assertEqual(hash(true), 258181728628715636)
        self.assertEqual(hash(Vector(1.0)), 258181728628715636)
        self.assertEqual(hash(Vector("Hello world!")), 2211555187250521325)
        self.assertEqual(hash(Vector(["Hello ", "world!"])), -463109597642348796)
        self.assertEqual(hash(Vector(["foo", 1])), hash(Vector(["foo", 1.0])))
        self.assertNotEqual(hash(Vector(["foo", 1.0])), hash(Vector(["foo", 1.1])))
        self.assertEqual(hash(Vector(Node('foo', {'bar'}, {'test': Vector(5)}))), 9166098286934782834)
        self.assertEqual(hash(Vector.symbol('foo')), hash(Vector(FOO_SYMBOL_NUMBER)))
        self.assertIsNotNone(hash(Vector(test_func)))  # just check it works, value will not be stable
        self.assertIsNotNone(hash(Vector(test_class)))  # just check it works, value will not be stable

    def test_hash_uniformity(self):
        from scipy.stats import kstest
        hashes = []
        scale = 1 << 64
        for c0 in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            for c1 in 'abcdefghijklmnopqrstuvwxyz':
                for i in range(100):
                    hashes.append((hash(Vector([c0 + c1, i])) % scale) / scale)
        result = kstest(hashes, 'uniform')
        self.assertGreater(result.pvalue, 0.05)

    def test_match(self):
        self.assertEqual(null.match(1, float), None)
        self.assertEqual(null.match(1, int), None)
        self.assertEqual(null.match(1, str), None)
        self.assertEqual(null.match(1, float, 3.5), 3.5)
        self.assertEqual(null.match(3, float, [1, 2, 3.5]), [1, 2, 3.5])
        self.assertEqual(Vector(3.5).match(0, float), [3.5])
        self.assertEqual(Vector(3.5).match(1, float), 3.5)
        self.assertEqual(Vector(3.5).match(1, int), 3)
        self.assertEqual(Vector(-3.5).match(1, int), -4)
        self.assertEqual(Vector(3.5).match(1, str), None)
        self.assertEqual(Vector(3.5).match(3, float), [3.5, 3.5, 3.5])
        self.assertEqual(Vector("Hello world!").match(1, str), "Hello world!")
        self.assertEqual(Vector("Hello world!").match(2, str), ["Hello world!", "Hello world!"])
        self.assertEqual(Vector(["Hello ", "world!"]).match(2, str), ["Hello ", "world!"])
        self.assertEqual(Vector(["Hello ", 3.5]).match(2, str), ["Hello ", "3.5"])
        self.assertEqual(Vector(["Hello ", "world!"]).match(1, str), None)
        self.assertEqual(true.match(1, float), 1.0)
        self.assertIs(true.match(1, bool), True)
        self.assertEqual(true.match(2, bool), [True, True])
        self.assertEqual(Vector([0, 1]).match(2, bool), [False, True])
        self.assertEqual(Vector.symbol('foo').match(1, float), FOO_SYMBOL_NUMBER)
        self.assertEqual(Vector.symbol('foo').match(1, str), 'foo')
        self.assertEqual(Vector.symbol('foo').match(2, str), ['foo', 'foo'])
        self.assertEqual(Vector.symbol('foo').concat(Vector.symbol('bar')).match(2, str), ['foo', 'bar'])

    def test_repr(self):
        self.assertEqual(repr(null), "null")
        self.assertEqual(repr(true), "1")
        self.assertEqual(repr(false), "0")
        self.assertEqual(repr(Vector(1 / 3)), "0.333333333")
        self.assertEqual(repr(Vector([1, 2, 3])), "1;2;3")
        self.assertEqual(repr(Vector([1, 2.5, 3])), "1;2.5;3")
        self.assertEqual(repr(Vector("Hello world!")), "'Hello world!'")
        self.assertEqual(repr(Vector("")), "''")
        self.assertEqual(repr(Vector([99, "red", "balloons"])), "99;'red';'balloons'")
        node = Node('foo', {'bar'}, {'color': Vector(1)})
        self.assertEqual(repr(Vector(node)), f"({node!r})")
        self.assertEqual(repr(Vector.symbol('foo')), ':foo')
        self.assertEqual(repr(Vector.symbol('foo').concat(Vector(99))), ':foo;99')

    def test_neg(self):
        self.assertEqual(-null, null)
        self.assertEqual(-Vector("Hello world!"), null)
        self.assertEqual(-Vector(-3), Vector(3))
        self.assertEqual(-Vector([0, 0.1, 3, -99, 1e99, math.inf]), Vector([0, -0.1, -3, 99, -1e99, -math.inf]))

    def test_pos(self):
        self.assertEqual(+null, null)
        self.assertEqual(+Vector("Hello world!"), null)
        self.assertEqual(+Vector(-3), Vector(-3))
        self.assertEqual(+Vector([0, 0.1, 3, -99, 1e99, math.inf]), Vector([0, 0.1, 3, -99, 1e99, math.inf]))
        self.assertEqual(+Vector.symbol('foo'), Vector.symbol('foo'))

    def test_abs(self):
        self.assertEqual(abs(null), null)
        self.assertEqual(abs(Vector("Hello world!")), null)
        self.assertEqual(abs(Vector(-3)), Vector(3))
        self.assertEqual(abs(Vector([0, 0.1, 3, -99, 1e99, math.inf])), Vector([0, 0.1, 3, 99, 1e99, math.inf]))

    def test_ceil(self):
        self.assertEqual(math.ceil(null), null)
        self.assertEqual(math.ceil(Vector("Hello world!")), null)
        self.assertEqual(math.ceil(Vector(-3.5)), Vector(-3))
        self.assertEqual(math.ceil(Vector([0, 0.1, 3.5, -99.5, 1e-99, math.inf])), Vector([0, 1, 4, -99, 1, math.inf]))

    def test_floor(self):
        self.assertEqual(math.floor(null), null)
        self.assertEqual(math.floor(Vector("Hello world!")), null)
        self.assertEqual(math.floor(Vector(-3.5)), Vector(-4))
        self.assertEqual(math.floor(Vector([0, 0.1, 3.5, -99.5, 1e-99, math.inf])), Vector([0, 0, 3, -100, 0, math.inf]))

    def test_fract(self):
        self.assertEqual(null.fract(), null)
        self.assertEqual(Vector("Hello world!").fract(), null)
        self.assertEqual(Vector(-3.5).fract(), Vector(0.5))
        self.assertTrue(all_isclose(Vector([0, 0.1, 3.5, -99.5, 1e-99, math.inf]).fract(), Vector([0, 0.1, 0.5, 0.5, 1e-99, math.nan])))

    def test_add(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x + null, null)
        self.assertEqual(null + x, null)
        self.assertEqual(x + Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") + x, null)
        self.assertEqual(x + Vector(1), Vector([2, 1.1, -4, 1000001, math.inf]))
        self.assertEqual(Vector(1) + x, Vector([2, 1.1, -4, 1000001, math.inf]))
        self.assertEqual(x + Vector([1, 2]), Vector([2, 2.1, -4, 1000002, math.inf]))
        self.assertEqual(Vector([1, 2]) + x, Vector([2, 2.1, -4, 1000002, math.inf]))
        self.assertEqual(x + x, Vector([2, 0.2, -10, 2e6, math.inf]))

    def test_mul_add(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x.mul_add(Vector(1), null), null)
        self.assertEqual(x.mul_add(null, Vector(1)), null)
        self.assertEqual(null.mul_add(x, Vector(1)), null)
        self.assertEqual(x.mul_add(Vector(1), Vector("Hello world!")), null)
        self.assertEqual(Vector("Hello world!").mul_add(x, Vector(1)), null)
        self.assertEqual(x.mul_add(Vector(0), Vector(1)), x)
        self.assertEqual(x.mul_add(Vector(1), Vector(1)), x + 1)
        self.assertEqual(x.mul_add(x, Vector(1)), x + x)
        self.assertEqual(Vector(0).mul_add(x, Vector(1)), x)
        self.assertEqual(Vector(1).mul_add(x, Vector(1)), x + 1)
        self.assertEqual(Vector(0).mul_add(x, x), x * x)
        self.assertEqual(Vector(x).mul_add(x, x), x * x + x)

    def test_sub(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x - null, null)
        self.assertEqual(null - x, null)
        self.assertEqual(x - Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") - x, null)
        self.assertEqual(x - Vector(1), Vector([0, -0.9, -6, 999999, math.inf]))
        self.assertEqual(Vector(1) - x, Vector([0, 0.9, 6, -999999, -math.inf]))
        self.assertEqual(x - Vector([1, 2]), Vector([0, -1.9, -6, 999998, math.inf]))
        self.assertEqual(Vector([1, 2]) - x, Vector([0, 1.9, 6, -999998, -math.inf]))
        y = x - x
        self.assertEqual(y[:4], Vector([0, 0, 0, 0]))
        self.assertTrue(math.isnan(y[4]))

    def test_mul(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x * null, null)
        self.assertEqual(null * x, null)
        self.assertEqual(x * Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") * x, null)
        self.assertEqual(x * Vector(1), x)
        self.assertEqual(Vector(1) * x, x)
        self.assertEqual(x * Vector([1, 2]), Vector([1, 0.2, -5, 2e6, math.inf]))
        self.assertEqual(Vector([1, 2]) * x, Vector([1, 0.2, -5, 2e6, math.inf]))
        self.assertTrue(all_isclose(x * x, Vector([1, 0.01, 25, 1e12, math.inf])))

    def test_truediv(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x / null, null)
        self.assertEqual(null / x, null)
        self.assertEqual(x / Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") / x, null)
        self.assertEqual(x / Vector(1), x)
        self.assertEqual(Vector(1) / x, Vector([1, 10, -0.2, 1e-6, 0]))
        self.assertEqual(x / Vector([1, 2]), Vector([1, 0.05, -5, 5e5, math.inf]))
        self.assertEqual(Vector([1, 2]) / x, Vector([1, 20, -0.2, 2e-6, 0]))
        y = x / x
        self.assertEqual(y[:4], Vector([1, 1, 1, 1]))
        self.assertTrue(math.isnan(y[4]))

    def test_floordiv(self):
        x = Vector([1, 0.1, -5.3, 1e6, math.inf])
        self.assertEqual(x // null, null)
        self.assertEqual(null // x, null)
        self.assertEqual(x // Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") // x, null)
        self.assertEqual(x // Vector(1), Vector([1, 0, -6, 1e6, math.inf]))
        self.assertEqual(Vector(1) // x, Vector([1, 10, -1, 0, 0]))
        self.assertEqual(x // Vector([1, 2]), Vector([1, 0, -6, 5e5, math.inf]))
        self.assertEqual(Vector([1, 2]) // x, Vector([1, 20, -1, 0, 0]))
        y = x // x
        self.assertEqual(y[:4], Vector([1, 1, 1, 1]))
        self.assertTrue(math.isnan(y[4]))

    def test_mod(self):
        x = Vector([1, 0.1, -5, 1e6])
        self.assertEqual(x % null, null)
        self.assertEqual(null % x, null)
        self.assertEqual(x % Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") % x, null)
        self.assertEqual(x % Vector(2), Vector([1, 0.1, 1, 0]))
        self.assertEqual(Vector(2) % x, Vector([0, 0, -3, 2]))
        self.assertEqual(x % Vector([2, 3]), Vector([1, 0.1, 1, 1]))
        self.assertEqual(Vector([2, 3]) % x, Vector([0, 0, -3, 3]))
        self.assertEqual(x % x, Vector([0, 0, 0, 0]))

    def test_pow(self):
        x = Vector([1, 0.1, -5, 1e6, math.inf])
        self.assertEqual(x ** null, null)
        self.assertEqual(null ** x, null)
        self.assertEqual(x ** Vector("Hello world!"), null)
        self.assertEqual(Vector("Hello world!") ** x, null)
        self.assertTrue(all_isclose(x ** Vector(2), Vector([1, 0.01, 25, 1e12, math.inf])))
        self.assertTrue(all_isclose(Vector(2) ** x, Vector([2, 1.0717734625362931, 0.03125, math.inf, math.inf])))
        self.assertTrue(all_isclose(x ** Vector([1, 2]), Vector([1, 0.01, -5, 1e12, math.inf])))
        self.assertTrue(all_isclose(Vector([1, 2]) ** x, Vector([1, 1.0717734625362931, 1, math.inf, 1])))
        self.assertTrue(all_isclose(x * x, Vector([1, 0.01, 25, 1e12, math.inf])))

    def test_eq(self):
        self.assertTrue(null == Vector())
        self.assertTrue(true == Vector(1))
        self.assertTrue(false == Vector(0))
        self.assertTrue(Vector([1, 2, 3]) == Vector([1, 2, 3]))
        self.assertTrue(Vector(["Hello ", "world!"]) == Vector(["Hello ", "world!"]))
        self.assertTrue(Vector.symbol('foo') == Vector(FOO_SYMBOL_NUMBER))
        self.assertTrue(Vector.symbol('foo') == Vector.symbol('foo'))

    def test_ne(self):
        self.assertFalse(null != Vector())
        self.assertTrue(true != Vector(0))
        self.assertTrue(false != Vector(1))
        self.assertFalse(Vector([1, 2, 3]) != Vector([1, 2, 3]))
        self.assertTrue(Vector(["Hello ", "world!"]) != Vector(["Hello world!"]))
        self.assertTrue(Vector.symbol('foo') != Vector.symbol('bar'))

    def test_gt(self):
        self.assertTrue(Vector(1) > null)
        self.assertTrue(true > Vector(0))
        self.assertFalse(false > Vector(1))
        self.assertTrue(Vector([1, 2, 4]) > Vector([1, 2, 3]))
        self.assertTrue(Vector([1, 3]) > Vector([1, 2, 3]))
        self.assertTrue(Vector(["Hello world!"]) > Vector(["Hello ", "world!"]))
        self.assertTrue(Vector(["Z"]) > Vector(["Hello world!"]))

    def test_ge(self):
        self.assertTrue(Vector(1) >= null)
        self.assertTrue(true >= Vector(0))
        self.assertTrue(false >= Vector(0))
        self.assertFalse(Vector([1, 2, 3]) >= Vector([1, 2, 3, 4]))
        self.assertTrue(Vector([1, 2, 3]) >= Vector([1, 2, 3]))
        self.assertTrue(Vector([1, 3]) >= Vector([1, 2, 3]))
        self.assertTrue(Vector(["Hello world!"]) >= Vector(["Hello ", "world!"]))
        self.assertTrue(Vector(["Z"]) >= Vector(["Hello world!"]))

    def test_lt(self):
        self.assertTrue(null < Vector(1))
        self.assertTrue(Vector(0) < true)
        self.assertFalse(Vector(1) < false)
        self.assertTrue(Vector([1, 2, 3]) < Vector([1, 2, 4]))
        self.assertTrue(Vector([1, 2, 3]) < Vector([1, 3]))
        self.assertTrue(Vector(["Hello ", "world!"]) < Vector(["Hello world!"]))
        self.assertTrue(Vector(["Hello world!"]) < Vector(["Z"]))

    def test_le(self):
        self.assertTrue(null <= Vector(1))
        self.assertTrue(Vector(0) <= true)
        self.assertTrue(Vector(0) <= false)
        self.assertFalse(Vector([1, 2, 3, 4]) <= Vector([1, 2, 3]))
        self.assertTrue(Vector([1, 2, 3]) <= Vector([1, 2, 3]))
        self.assertTrue(Vector([1, 2, 3]) <= Vector([1, 3]))
        self.assertTrue(Vector(["Hello ", "world!"]) <= Vector(["Hello world!"]))
        self.assertTrue(Vector(["Hello world!"]) <= Vector(["Z"]))

    def test_getitem(self):
        x = Vector.range(100)
        self.assertTrue(x[:10] == Vector.range(10))
        self.assertTrue(x[10:100] == Vector.range(10, 100))
        self.assertTrue(x[10:100:5] == Vector.range(10, 100, 5))
        self.assertTrue(x[99:9:-1] == Vector.range(99, 9, -1))
        self.assertTrue(x[99:9:-2] == Vector.range(99, 9, -2))
        self.assertTrue(x[99:9:-13 / 3] == Vector([99, 94, 90, 86, 81, 77, 73, 68, 64, 60, 55, 51, 47, 42, 38, 34, 29, 25, 21, 16, 12]))

    def test_normalize(self):
        self.assertEqual(null.normalize(), null)
        self.assertEqual(Vector(["Hello world!"]).normalize(), null)
        self.assertEqual(Vector(1).normalize(), Vector(1))
        self.assertEqual(Vector([0, 1]).normalize(), Vector([0, 1]))
        self.assertTrue(all_isclose(Vector([1, 1]).normalize(), Vector([0.707106781, 0.707106781])))
        self.assertTrue(all_isclose(Vector([1, -2, 3]).normalize(), Vector([0.267261242, -0.534522484, 0.801783726])))

    def test_dot(self):
        self.assertEqual(null.dot(null), null)
        self.assertEqual(null.dot(Vector([1, 2, 3])), null)
        self.assertEqual(Vector([1, 2, 3]).dot(null), null)
        self.assertEqual(Vector([1, 2, 3]).dot(Vector([1, 2, 3])), Vector(14))
        self.assertEqual(Vector([1, 2, 3]).dot(Vector([3, 2, 1])), Vector(10))
        self.assertEqual(Vector([1, 2, 3]).dot(Vector(3)), Vector(18))
        self.assertEqual(Vector([1, 2, 3]).dot(Vector([3, 2])), Vector(16))

    def test_cross(self):
        self.assertEqual(null.cross(null), null)
        self.assertEqual(null.cross(Vector([1, 2, 3])), null)
        self.assertEqual(Vector([1, 2, 3]).cross(null), null)
        self.assertEqual(Vector([1, 2, 3]).cross(Vector([1, 2, 3])), Vector([0, 0, 0]))
        self.assertEqual(Vector([1, 2, 3]).cross(Vector([-1, -2, -3])), Vector([0, 0, 0]))
        self.assertEqual(Vector([1, 2, 3]).cross(Vector([1, 2])), null)
        self.assertEqual(Vector([1, 2, 3]).cross(Vector([3, 2, 1])), Vector([-4, 8, -4]))

    def test_concat(self):
        a = Vector([1, 2, 3])
        b = Vector([4, 5, 6])
        self.assertIs(a.concat(null), a)
        self.assertIs(null.concat(b), b)
        self.assertEqual(a.concat(b), Vector([1, 2, 3, 4, 5, 6]))
        self.assertEqual(b.concat(a), Vector([4, 5, 6, 1, 2, 3]))
        self.assertTrue(a.concat(b).numeric)
        self.assertEqual(Vector(['hello']).concat(b), Vector(['hello', 4, 5, 6]))
        self.assertEqual(a.concat(Vector(['world'])), Vector([1, 2, 3, 'world']))
        self.assertEqual(Vector(['hello']).concat(Vector(['world'])), Vector(['hello', 'world']))



"""
Flitter generic MIDI interface
"""

import asyncio
from dataclasses import dataclass
import enum

from loguru import logger
import rtmidi2

from ...clock import system_clock


class MIDI(enum.IntEnum):
    NOTE_OFF = 0x80
    NOTE_ON = 0x90
    POLYPHONIC_PRESSURE = 0xA0
    CONTROL_CHANGE = 0xB0
    PROGRAM_CHANGE = 0xC0
    CHANNEL_PRESSURE = 0xD0
    PITCH_BEND_CHANGE = 0xE0
    SYSEX_START = 0xF0
    SYSEX_END = 0xF7
    CLOCK = 0xF8
    START = 0xFA
    CONTINUE = 0xFB
    STOP = 0xFC
    ACTIVE = 0xFE
    RESET = 0xFF


@dataclass
class MidiEvent:
    timestamp: float
    channel: int


@dataclass
class NoteEvent(MidiEvent):
    note: int


@dataclass
class NoteOnEvent(NoteEvent):
    velocity: int


@dataclass
class NotePressureEvent(NoteEvent):
    pressure: int


@dataclass
class NoteOffEvent(NoteEvent):
    pass


@dataclass
class ControlChangeEvent(MidiEvent):
    control: int
    value: int


@dataclass
class ChannelPressureEvent(MidiEvent):
    pressure: int


@dataclass
class PitchBendEvent(MidiEvent):
    value: int


class MidiPort:
    def __init__(self, name, input=True, output=True, sysex_id=None, virtual=False):
        self.name = name
        self.sysex_id = bytes(sysex_id) if sysex_id is not None else None
        self._midi_in = None
        self._midi_out = None
        try:
            if input:
                self._midi_in = rtmidi2.MidiIn()
                if virtual:
                    self._midi_in.open_virtual_port(name)
                else:
                    for port in rtmidi2.get_in_ports():
                        if port.split(':')[0] == name:
                            name = port
                    self._midi_in.open_port(name)
                self._last_receive_timestamp = None
                self._loop = asyncio.get_event_loop()
                self._receive_queue = asyncio.Queue()
                self._midi_in.callback = self._message_received
            if output:
                self._midi_out = rtmidi2.MidiOut()
                if virtual:
                    self._midi_out.open_virtual_port(self.name)
                else:
                    for port in rtmidi2.get_in_ports():
                        if port.split(':')[0] == name:
                            name = port
                    self._midi_out.open_port(name)
        except RuntimeError as exc:
            if self._midi_in is not None:
                self._midi_in.close()
            raise ValueError(f"Unable to open port {name}") from exc
        logger.debug("Opened MIDI port '{}'", self.name)

    def close(self):
        if self._midi_in is not None:
            self._midi_in.close_port()
            self._midi_in = None
        if self._midi_out is not None:
            self._midi_out.close_port()
            self._midi_out = None

    def _message_received(self, message, delta):
        now = system_clock()
        if self._last_receive_timestamp is None:
            timestamp = now
        else:
            timestamp = min(now, self._last_receive_timestamp + delta)
        self._last_receive_timestamp = timestamp
        logger.trace("Received MIDI @ {:.2f} - {}", timestamp, " ".join(f"{b:02x}" for b in message))
        self._loop.call_soon_threadsafe(self._receive_queue.put_nowait, (message, timestamp))

    async def wait_event(self, timeout=None):
        if self._midi_in is None:
            raise TypeError("MIDI port not open for input")
        try:
            message, timestamp = await asyncio.wait_for(self._receive_queue.get(), timeout=timeout)
        except asyncio.TimeoutError:
            return None
        event_type = message[0] & 0xf0
        channel = message[0] & 0x0f
        if event_type == MIDI.NOTE_ON:
            return NoteOnEvent(note=message[1], velocity=message[2], channel=channel, timestamp=timestamp)
        elif event_type == MIDI.POLYPHONIC_PRESSURE:
            return NotePressureEvent(note=message[1], pressure=message[2], channel=channel, timestamp=timestamp)
        elif event_type == MIDI.NOTE_OFF:
            return NoteOffEvent(note=message[1], channel=channel, timestamp=timestamp)
        elif event_type == MIDI.CONTROL_CHANGE:
            return ControlChangeEvent(control=message[1], value=message[2], channel=channel, timestamp=timestamp)
        elif event_type == MIDI.CHANNEL_PRESSURE:
            return ChannelPressureEvent(pressure=message[1], channel=channel, timestamp=timestamp)
        elif event_type == MIDI.PITCH_BEND_CHANGE:
            return PitchBendEvent(value=((message[2] << 7) + message[1]), channel=channel, timestamp=timestamp)
        logger.debug("Unhandled MIDI message: {}", " ".join(f"{b:02x}" for b in message))
        return None

    def send_raw(self, *data):
        if self._midi_out is None:
            raise TypeError("MIDI port not open for output")
        logger.trace("Send MIDI - {}", " ".join(f"{b:02x}" for b in data))
        self._midi_out.send_raw(*data)

    def send_note_off(self, note, channel=0):
        self.send_raw(MIDI.NOTE_OFF + channel, note)

    def send_note_on(self, note, velocity=127, channel=0):
        self.send_raw(MIDI.NOTE_ON + channel, note, velocity)

    def send_note_pressure(self, note, pressure, channel=0):
        self.send_raw(MIDI.POLYPHONIC_PRESSURE + channel, note, pressure)

    def send_channel_pressure(self, pressure, channel=0):
        self.send_raw(MIDI.CHANNEL_PRESSURE + channel, pressure)

    def send_control_change(self, control, value, channel=0):
        self.send_raw(MIDI.CONTROL_CHANGE + channel, control, value)

    def send_program_change(self, program, channel=0):
        self.send_raw(MIDI.PROGRAM_CHANGE + channel, program)

    def send_pitch_bend_change(self, value, channel=0):
        self.send_raw(MIDI.PITCH_BEND_CHANGE + channel, value & 0x7f, value >> 7)

    def send_sysex(self, *data, sysex_id=None):
        if sysex_id is None:
            sysex_id = self.sysex_id
            if sysex_id is None:
                raise TypeError("MIDI port sysex_id not set")
        self.send_raw(MIDI.SYSEX_START, *sysex_id, *data, MIDI.SYSEX_END)

    def send_clock(self):
        self.send_raw(MIDI.CLOCK)

    def send_start(self):
        self.send_raw(MIDI.START)

    def send_continue(self):
        self.send_raw(MIDI.CONTINUE)

    def send_stop(self):
        self.send_raw(MIDI.STOP)

    def send_active(self):
        self.send_raw(MIDI.ACTIVE)

    def send_reset(self):
        self.send_raw(MIDI.RESET)



# cython: language_level=3, profile=True

"""
Flitter DMX control
"""

import asyncio
import enum
import struct

import cython
from libc.math cimport round
from loguru import logger

from .. import name_patch
from ..clock import system_clock
from .. cimport model
from ..streams import SerialStream


logger = name_patch(logger, __name__)


class DMXDriver:
    async def connect(self, tty=None):
        raise NotImplementedError()

    async def send_dmx_data(self, data):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class EntecDMXDriver(DMXDriver):
    DEFAULT_BAUD_RATE = 57600
    PACKET_START = 0x7e
    PACKET_END = 0xe7

    class Label(enum.IntEnum):
        GetParameters = 3
        SetParameters = 4
        ReceivedDMXPacket = 5
        SendDMXPacket = 6

    class Parameters:
        @classmethod
        def from_payload(cls, payload):
            firmware_version, break_time, mark_after_break_time, refresh_rate = struct.unpack_from('<HBBB', payload)
            user_data = payload[6:]
            return cls(firmware_version, break_time, mark_after_break_time, refresh_rate, user_data)

        def __init__(self, firmware_version=None, break_time=9, mark_after_break_time=1, refresh_rate=40, user_data=b''):
            self.firmware_version = firmware_version
            self.break_time = break_time
            self.mark_after_break_time = mark_after_break_time
            self.refresh_rate = refresh_rate
            self.user_data = user_data

        def __bytes__(self):
            return struct.pack('<HBBB', len(self.user_data), self.break_time, self.mark_after_break_time, self.refresh_rate)

    def __init__(self, timeout=2.5):
        self._timeout = timeout
        self._stream = None
        self.parameters = None

    async def connect(self, tty=None, baudrate=None):
        try:
            if baudrate is None:
                baudrate = self.DefaultBaudrate
            if tty is not None:
                self._stream = SerialStream(tty, baudrate=baudrate)
            else:
                self._stream = SerialStream.stream_matching(vid=0x0403, pid=0x6001, baudrate=baudrate)
            await self.send_packet(self.Label.GetParameters, struct.pack('<H', 508))
            label, payload = await self.recv_packet()
            if label != self.Label.GetParameters:
                raise ConnectionError("Label mismatch on initial parameters request")
            self.parameters = self.Parameters.from_payload(payload)
            logger.info("Connected to DMX interface; firmware version {}, refresh rate {}", self.parameters.firmware_version, self.parameters.refresh_rate)
        except ConnectionError:
            raise
        except Exception as exc:
            raise ConnectionError("Unable to connect") from exc

    async def send_packet(self, label, payload, pad=None):
        length = len(payload) if pad is None else max(len(payload), pad)
        data = bytearray(length + 5)
        data[:4] = struct.pack('<BBH', self.PACKET_START, label, length)
        data[4:4+len(payload)] = payload
        data[-1] = self.PACKET_END
        self._stream.write(data)

    async def _readexactly(self, nbytes, until_time):
        return await asyncio.wait_for(self._stream.readexactly(nbytes), max(0, until_time - system_clock()))

    async def recv_packet(self):
        fail_time = system_clock() + self._timeout
        data = bytearray()
        junk = 0
        while True:
            data += await self._readexactly(4 - len(data), fail_time)
            index = data.find(self.PACKET_START)
            if index == 0:
                break
            if index == -1:
                index = len(data)
            junk += index
            del data[:index]
        if junk:
            logger.warning("Discarded {} leading bytes of junk in stream", junk)
        _, label, payload_length = struct.unpack('<BBH', data)
        payload = await self._readexactly(payload_length, fail_time)
        assert (await self._readexactly(1, fail_time)) == bytes([self.PACKET_END])
        return label, payload

    async def send_dmx_data(self, data):
        await self.send_packet(self.Label.SendDMXPacket, data, pad=25)

    def close(self):
        self._stream.close()
        self._stream = None


@cython.boundscheck(False)
@cython.wraparound(False)
cdef frame_and_escape(unsigned char[:] data):
    cdef int n=0, i, j
    cdef unsigned char b
    for i in range(len(data)):
        b = data[i]
        if b == 0x7d or b == 0x7e:
            n += 2
        else:
            n += 1
    cdef unsigned char[:] frame = bytearray(n + 2)
    frame[0] = 0x7e
    j = 1
    for i in range(len(data)):
        b = data[i]
        if b == 0x7d or b == 0x7e:
            frame[j] = 0x7d
            frame[j+1] = b ^ 0x20
            j += 2
        else:
            frame[j] = b
            j += 1
    frame[j] = 0x7e
    return frame


class OutputArtsDMXDriver(DMXDriver):
    DEFAULT_BAUD_RATE = 115200

    def __init__(self):
        self._stream = None

    async def connect(self, tty=None, baudrate=None):
        if tty is not None:
            self._stream = SerialStream(tty, baudrate=baudrate)
        else:
            self._stream = SerialStream.stream_matching(vid=0x03eb, pid=0x2044, baudrate=baudrate)

    async def send_dmx_data(self, data):
        frame = frame_and_escape(data)
        self._stream.write(bytes(frame))
        await self._stream.drain()

    def close(self):
        if self._stream is not None:
            self._stream.close()
            self._stream = None


cdef class DMX:
    cdef object driver

    def __init__(self, **kwargs):
        self.driver = None

    def purge(self):
        pass

    def destroy(self):
        if self.driver is not None:
            self.driver.close()
            self.driver = None

    async def update(self, engine, model.Node node, **kwargs):
        driver = node.get('driver', 1, str, '').lower()
        cls = {'entec': EntecDMXDriver, 'outputarts': OutputArtsDMXDriver}.get(driver)
        if cls is not None:
            tty = node.get('id', 1, str)
            baudrate = node.get('baudrate', 1, int, cls.DEFAULT_BAUD_RATE)
            if not isinstance(self.driver, cls):
                if self.driver is not None:
                    self.driver.close()
                self.driver = cls()
                await self.driver.connect(tty, baudrate=baudrate)
            channel_data = bytearray(513)
            end = self.collect_channels(node, channel_data)
            await self.driver.send_dmx_data(channel_data[:end])
        elif self.driver is not None:
            self.driver.close()
            self.driver = None

    cpdef int collect_channels(self, model.Node node, unsigned char[:] channel_data):
        cdef model.Node child
        cdef double channel
        cdef list channels
        cdef int start, end, i, n=1
        for child in node.children:
            if child.kind == 'fixture':
                address = child.get('address', 1, int)
                channels = child.get('channels', 0, float)
                if address is not None and channels is not None:
                    start = int(address)
                    end = start + len(channels)
                    if start >=1 and end <= len(channel_data):
                        for i, channel in zip(range(start, end), channels):
                            channel_data[i] = min(max(0, <int>(round(channel*255))), 255)
                        if end > n:
                            n = end
        return n


RENDERER_CLASS = DMX



"""
Video scene node
"""

from loguru import logger
import PIL.Image

from . import SceneNode
from ...cache import SharedCache
from .glconstants import GL_SRGB8


class Image(SceneNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._filename = None
        self._image = None
        self._texture = None

    def release(self):
        self._filename = None
        self._image = None
        self._texture = None

    @property
    def texture(self):
        return self._texture

    async def update(self, engine, node, default_size=(512, 512), **kwargs):
        references = kwargs.setdefault('references', {})
        if node_id := node.get('id', 1, str):
            references[node_id] = self
        self.hidden = node.get('hidden', 1, bool, False)
        self.tags = node.tags
        filename = node.get('filename', 1, str)
        if self._image is not None:
            default_size = [self._image.width, self._image.height]
        else:
            default_size = None
        size = node.get('size', 2, int, default_size)
        if filename != self._filename or (size and size != [self.width, self.height]):
            self.release()
            self._filename = filename
            self.width = None
            self.height = None
        if self._filename is not None:
            image = SharedCache[filename].read_pil_image()
            if image is not self._image:
                self._image = image
                if size and size != [image.width, image.height]:
                    image = image.resize(size, PIL.Image.Resampling.BILINEAR)
                    logger.debug("Resized {} to {}x{}", self._filename, *size)
                self.width = image.width
                self.height = image.height
                flipped = image.transpose(PIL.Image.FLIP_TOP_BOTTOM)
                if self._image.has_transparency_data:
                    self._texture = self.glctx.texture((self.width, self.height), 4, internal_format=GL_SRGB8)
                    self._texture.write(flipped.convert('RGBA').convert('RGBa').tobytes())
                else:
                    self._texture = self.glctx.texture((self.width, self.height), 3, internal_format=GL_SRGB8)
                    self._texture.write(flipped.convert('RGB').tobytes())

    def similar_to(self, node):
        return super().similar_to(node) and node.get('filename', 1, str) == self._filename


SCENE_NODE_CLASS = Image
